import networkx as nx
import pytest

from geco.graphs.utilities import *


def test_edge_properties():
    num_nodes = 7
    graph = nx.path_graph(num_nodes)
    for u, v, data in graph.edges(data=True):
        data["weight"] = u

    max_weight, min_weight, num_of_zero_weighted = edgeweight_properties(graph)
    assert max_weight == num_nodes - 2
    assert min_weight == 0
    assert num_of_zero_weighted == 1


@pytest.mark.parametrize("n", [3, 10, 100])
def test_graph_properties(n):
    graph = nx.path_graph(n)
    for u, v, data in graph.edges(data=True):
        data["weight"] = u

    properties = graph_properties(graph)
    assert properties["num_nodes"] == n
    assert properties["num_edges"] == n - 1
    assert properties["avg_degree"] == 2 * (n - 1) / n
    assert properties["density"] == 2 * (n - 1) / (n * (n - 1))
    assert properties["planar"] == True
    assert properties["max_edgeweight"] == n - 2
    assert properties["min_edgeweight"] == 0
    assert properties["num_of_zero_edgeweights"] == 1
    assert properties["num_of_connected_components"] == 1
    assert properties["max_degree"] == 2
    assert -1.0 <= properties["assortativity_coeff"] <= -0.010204081632620635
    assert properties["number_of_triangles"] == 0
    assert properties["max_k_core"] == 1
    assert properties["average_clustering_coeff"] == 0



from geco.mips.scheduling.generic import *


@py_random_state(-1)
def heinz_params(number_of_facilities, number_of_tasks, seed=0):
    """Generates scheduling MIP instance params according to [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int,int]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility

    References
    ----------
    .. [1] Heinz, J. (2013). Recent Improvements Using Constraint Integer Programming for Resource Allocation and Scheduling.
    In Integration of AI and OR Techniques in Constraint Programming for Combinatorial Optimization Problems
    (pp. 12–27). Springer Berlin Heidelberg.
    """
    return generate_params(number_of_facilities, number_of_tasks, seed)


@py_random_state(-1)
def heinz_instance(number_of_facilities, number_of_tasks, seed=0):
    """Generates scheduling MIP instance according to [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
        model: SCIP model of generated instance

    References
    ----------
    .. [1] Heinz, J. (2013). Recent Improvements Using Constraint Integer Programming for Resource Allocation and Scheduling.
    In Integration of AI and OR Techniques in Constraint Programming for Combinatorial Optimization Problems
    (pp. 12–27). Springer Berlin Heidelberg.
    """
    return heinz_formulation(
        number_of_facilities,
        number_of_tasks,
        *heinz_params(number_of_facilities, number_of_tasks, seed),
        name="Heinz Scheduling Instance",
    )



import pyscipopt as scip


def knapsack(weights, profits, capacity, name="Knapsack"):
    """Generates a knapsack MIP formulation.

    Parameters:
    ----------
        profits: list[float]
            List of profits of each item
        weights: list[float]
            List of weights of each item
        capacity: float
            Capacity of knapsack

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance
    """
    assert len(weights) == len(profits)
    assert capacity >= 0
    assert all(w >= 0 for w in weights)
    assert all(p >= 0 for p in profits)

    model = scip.Model(name)
    # add variables and their cost
    variables = [model.addVar(lb=0, ub=1, obj=profit, vtype="B") for profit in profits]
    # add constraints
    model.addCons(
        scip.quicksum(weight * variable for weight, variable in zip(weights, variables))
        <= capacity
    )

    model.setMaximize()

    return model



import networkx as nx
import pyscipopt as scip


def independent_set(graph, name="Independent Set"):
    """
    Generates an independent set instance according to [1].

    Parameters
    ----------
    graph: nx.Graph
        Networkx undirected graph
    name: str
        Name of the generated model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] https://www.princeton.edu/~aaa/Public/Teaching/ORF523/S16/ORF523_S16_Lec11_gh.pdf
    """
    graph = nx.relabel.convert_node_labels_to_integers(graph)
    model = scip.Model(name)

    vars = {
        str(node): model.addVar(lb=0, ub=1, obj=1, name=str(node), vtype="B")
        for node in graph.nodes
    }

    for u, v in graph.edges:
        model.addCons(vars[str(u)] + vars[str(v)] <= 1)

    model.setMaximize()

    return model


def _get_cliques(graph):
    """
    Partition the graph into cliques using a greedy algorithm, this code is
    based on the code from [1].

    Parameters
    ----------
    graph: nx.Graph
        Networkx undirected graph

    Returns
    -------
    cliques: list[set]
        The resulting clique partition

    References
    ----------
    .. [1] https://github.com/ds4dm/learn2branch/blob/master/01_generate_instances.py
    """
    cliques = []

    # sort nodes in descending order of degree
    leftover_nodes = sorted(list(graph.nodes), key=lambda node: -graph.degree[node])

    while leftover_nodes:
        clique_center, leftover_nodes = leftover_nodes[0], leftover_nodes[1:]
        clique = {clique_center}
        neighbors = set(graph.neighbors(clique_center)).intersection(leftover_nodes)
        densest_neighbors = sorted(neighbors, key=lambda node: -graph.degree[node])
        for neighbor in densest_neighbors:
            if all(
                [neighbor in graph.neighbors(clique_node) for clique_node in clique]
            ):
                clique.add(neighbor)
        cliques.append(clique)
        leftover_nodes = [node for node in leftover_nodes if node not in clique]

    return cliques


def clique_independent_set(graph, name="Clique Independent Set"):
    """
    Generates an independent set instance according to [1, 4.6.4].

    Parameters
    ----------
    graph: nx.Graph
        Networkx undirected graph
    name: str
        Name of the generated model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] David Bergman, Andre A. Cire, Willem-Jan Van Hoeve, and John Hooker. Decision diagrams
    for optimization. Springer, 2016.
    """
    graph = nx.relabel.convert_node_labels_to_integers(graph)
    model = scip.Model(name)

    cliques = _get_cliques(graph)

    vars = {
        str(node): model.addVar(lb=0, ub=1, obj=1, name=str(node), vtype="B")
        for node in graph.nodes
    }

    for clique in cliques:
        model.addCons(scip.quicksum(vars[str(node)] for node in clique) <= 1)

    model.setMaximize()

    return model



from networkx.utils import py_random_state


@py_random_state(-1)
def generate(generating_function, seed=0):
    """
    Convenience wrapper to generate multiple instances.

    Parameters
    ----------
    generating_function:
        A function that accepts a seed and returns an instance.
    seed: int, random state or None
        seed for randomization.

    Returns
    -------
     A pyscipopt model.
    """
    while True:
        yield generating_function(seed)


@py_random_state(-1)
def generate_n(generating_function, n, seed=0):
    """
    Convenience wrapper to generate n instances.

    Parameters
    ----------
    generating_function:
        A function that accepts a seed and returns an instance.
    n: int
        Number of instances to generate.
    seed: int, random state or None
        seed for randomization.

    Returns
    -------
     Tuple (instance_number, instance)
    """
    for i in range(n):
        yield i, generating_function(seed)


@py_random_state(-1)
def common_substructure_generator(
    instance_generation_function,
    backbone,
    expand_params_function,
    seed=0,
):
    """
    Generates instances that have common substructure

    Parameters
    ----------
    instance_generation_function:
        base function that defines MIP
    backbone:
        instance parameters of the common substructure
    expand_params_function:
        function to expand instance params given a backbone and a seed
    seed: int, random object or None
        for randomization

    Returns
    -------
        generator object
    """
    while True:
        yield instance_generation_function(*expand_params_function(backbone, seed=seed))



import itertools

import pytest

from geco.generator import *
from geco.mips.facility_location.cornuejols import cornuejols_instance
from geco.mips.set_cover.generic import *
from geco.mips.set_cover.sun import *


def test_generate():
    gen = generate(lambda seed: cornuejols_instance(10, 5, 2, seed))
    for model in itertools.islice(gen, 10):
        assert type(model) == scip.Model


def test_generate_n():
    n = 50
    gen = generate_n(lambda seed: cornuejols_instance(6, 3, 2, seed), n=n)
    all_models = []
    for i, model in gen:
        assert type(model) == scip.Model
        all_models.append(model)
    assert len(all_models) == n

    for model1, model2 in itertools.combinations(all_models, 2):
        obj_coeffs_1 = [v.getObj() for v in model1.getVars()]
        obj_coeffs_2 = [v.getObj() for v in model2.getVars()]
        assert obj_coeffs_1 != obj_coeffs_2


@pytest.mark.parametrize(
    "n,m,seed", itertools.product([10, 100, 200], [10, 100, 200], [0, 1, 1337, 53115])
)
def test_common_substructure_generator_set_cover(n, m, seed):
    gen = common_substructure_generator(
        instance_generation_function=set_cover,
        backbone=sun_params(n, m),
        expand_params_function=lambda backbone, seed: expand_sun_params(
            (n + 10, m), backbone, seed=seed
        ),
        seed=seed,
    )
    for model in itertools.islice(gen, 10):
        assert model.getNVars() == n + 10
        assert model.getNConss() == m
        assert model.getObjectiveSense() == "minimize"



import numpy as np
import networkx as nx
from networkx.utils import py_random_state, np_random_state
from numpy.random import random
import dwave_networkx as dwave


@py_random_state(-1)
def draw_intra_weight(seed=0):
    return seed.random() - 0.5


@py_random_state(-1)
def draw_inter_weight(seed=0):
    return seed.random() * 2 - 1


def chimera_graph(n, m, t, inter=lambda: 0, intra=lambda: 0):
    """
    Generate DWave Chimera graph as described in [1].

    Parameters
    ----------
    n: int
        Number of cells per row
    m: int
        Number of cells per column
    t: int
        Number of nodes on each side of a bipartite cell subgraph
    inter: function (seed) -> number
        Function to call for weights of inter-cell edges
    intra: function (seed) -> number
        Function to call for weights of intra-cell edges

    Returns
    -------
    graph: nx.Graph
        The generated Chimera graph

    References
    ----------
    ..[1] https://docs.ocean.dwavesys.com/en/latest/concepts/topology.html
    """
    graph = nx.Graph()
    # add nodes
    for cur_n in range(n):
        for cur_m in range(m):
            for cur_t in range(t):
                for cur_s in range(2):
                    graph.add_node((cur_n, cur_m, cur_t, cur_s))

    # add intra-cell edges
    for cur_n in range(n):
        for cur_m in range(m):
            for left_t in range(t):
                for right_t in range(t):
                    graph.add_edge(
                        (cur_n, cur_m, left_t, 0),
                        (cur_n, cur_m, right_t, 1),
                        weight=inter(),
                    )

    # add horizontal inter-cell edges
    for cur_n in range(n):
        for cur_m in range(m - 1):
            for cur_t in range(t):
                graph.add_edge(
                    (cur_n, cur_m, cur_t, 1),
                    (cur_n, cur_m + 1, cur_t, 1),
                    weight=intra(),
                )

    # add vertical inter-cell edges
    for cur_n in range(n - 1):
        for cur_m in range(m):
            for cur_t in range(t):
                graph.add_edge(
                    (cur_n, cur_m, cur_t, 0),
                    (cur_n + 1, cur_m, cur_t, 0),
                    weight=intra(),
                )

    return graph


"""
Parameterised graph generators
"""


@np_random_state(-1)
def selby_c(m, seed=0):
    """
    Generate Selby Chimera graph as described in section 7.3 in [1].

    Parameters
    ----------
    m: int
        Number of cells per column
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    graph: nx.Graph
        The generated Chimera graph

    References
    ----------
    ..[1] Jünger, M., Lobe, E., Mutzel, P., Reinelt, G., Rendl, F., Rinaldi, G., & Stollenwerk, T. (2019).
    Performance of a quantum annealer for Ising ground state computations on chimera graphs.
    arXiv preprint arXiv:1904.11965.
    """

    def inter_w():
        return seed.randint(low=-10, high=10 + 1) / 10.0

    def intra_w():
        return seed.randint(low=-5, high=5 + 1) / 10.0

    graph = chimera_graph(m, m, 4, inter=inter_w, intra=intra_w)

    graph.graph["name"] = f"selby_c{m}"
    graph.graph["seed"] = seed

    return graph


def _initialize_weights_chimera(
    chimera_graph, size, draw_inter_weight, draw_intra_weight, draw_other_weight
):
    c_coor = dwave.chimera_coordinates(size)
    for _from, _to in chimera_graph.edges:
        _from_nice = c_coor.linear_to_chimera(_from)
        _to_nice = c_coor.linear_to_chimera(_to)
        if in_same_chimera_tile(_from_nice, _to_nice):
            # edge from one side to the other (internal edge)
            if not on_same_side(_from_nice, _to_nice):
                chimera_graph.add_edge(_from, _to, weight=draw_intra_weight())
            else:  # odd couplers
                chimera_graph.add_edge(_from, _to, weight=draw_other_weight())
        else:
            chimera_graph.add_edge(_from, _to, weight=draw_inter_weight())


def on_same_side(_from_nice, _to_nice):
    u = 2
    return _from_nice[u] == _to_nice[u]


def in_same_chimera_tile(_from_nice, _to_nice):
    y, x = 0, 1
    return _from_nice[y] == _to_nice[y] and _from_nice[x] == _to_nice[x]


@py_random_state(-1)
def dwave_chimera_graph(
    m,
    n=None,
    t=4,
    draw_inter_weight=draw_inter_weight,
    draw_intra_weight=draw_intra_weight,
    draw_other_weight=draw_inter_weight,
    seed=0,
):
    """
    Generate DWave Chimera graph as described in [1] using dwave_networkx.

    Parameters
    ----------
    m: int
        Number of cells per column
    n: int
        Number of cells per row
    t: int
        Number of nodes on each side of a bipartite cell subgraph
    draw_inter_weight: function (seed) -> number
        Function to call for weights of inter-cell edges
    draw_intra_weight: function (seed) -> number
        Function to call for weights of intra-cell edges
    draw_other_weight: function (seed) -> number
            Function to call for weights of intra-cell edges
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    graph: nx.Graph
        The generated Chimera graph

    References
    ----------
    ..[1] https://docs.ocean.dwavesys.com/en/latest/concepts/topology.html
    """
    if not n:
        n = m
    g = dwave.chimera_graph(m, n, t)
    _initialize_weights_chimera(
        chimera_graph=g,
        size=m,
        draw_inter_weight=lambda: draw_inter_weight(seed),
        draw_intra_weight=lambda: draw_intra_weight(seed),
        draw_other_weight=lambda: draw_other_weight(seed),
    )
    return g


@py_random_state(-1)
def mgw(m=8, faulty=73, seed=0):
    """
    Generates McGeoch-Wang instances as described in [1].


    Parameters
    ----------
    m: int
        Chimera graph size parameter (produces mxm lattice)
    faulty: int
        Number of faulty nodes
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    graph: nx.Graph
        The generated Chimera graph

    References
    ----------
    ..[1] Michael Juenger, Elisabeth Lobe, Petra Mutzel, Gerhard Reinelt, Franz Rendl, Giovanni Rinaldi,
        & Tobias Stollenwerk. (2019). Performance of a Quantum Annealer for Ising Ground State Computations
        on Chimera Graphs.
    """

    def draw(seed):
        seed.choice((1, -1))

    graph = dwave_chimera_graph(
        m=m,
        draw_inter_weight=draw,
        draw_intra_weight=draw,
        draw_other_weight=draw,
        seed=seed,
    )
    faulty_nodes = set(seed.choices(range(graph.number_of_nodes()), k=faulty))
    to_be_removed_edges = set()
    for u, v in graph.edges():
        if u in faulty_nodes or v in faulty_nodes:
            to_be_removed_edges.add((u, v))
    graph.remove_edges_from(to_be_removed_edges)
    return graph



import math

import pytest

from geco.mips.scheduling.heinz import *
from geco.mips.scheduling.hooker import *


def test_hooker_formulation():
    params = number_of_facilities, number_of_tasks, time_steps = 3, 10, 100
    model = hooker_instance(*params)
    _check_hookers_instance(model, number_of_facilities, number_of_tasks, time_steps)


def test_hooker_generation():
    for params, model in generate_hookers_instances():
        _check_hookers_instance(model, *params[:-1])


@pytest.mark.parametrize(
    "number_of_facilities,number_of_tasks,seed",
    itertools.product([1, 2, 3], [5, 10, 15], [0, 1, 1337, 53115]),
)
def test_heinz_formulation(number_of_facilities, number_of_tasks, seed):
    main_params = number_of_facilities, number_of_tasks
    p, C, c, R, d, r = heinz_params(*main_params, seed)
    time_steps = int(max(d) - min(R))
    model = heinz_instance(*main_params)
    x_vars_count = number_of_facilities * number_of_tasks
    y_vars_lowerbound = 0
    y_vars_upperbound = number_of_facilities * number_of_tasks * time_steps
    assert (
        x_vars_count + y_vars_lowerbound
        <= model.getNVars()
        <= x_vars_count + y_vars_upperbound
    )
    constraints_lowerbound = (
        number_of_facilities
        + number_of_facilities * number_of_tasks
        + number_of_tasks
        + number_of_tasks
        + number_of_facilities * number_of_tasks
        + number_of_facilities * number_of_tasks
    )
    constraints_upperbound = (
        number_of_facilities
        + number_of_facilities * number_of_tasks
        + number_of_tasks * time_steps
        + number_of_tasks * (time_steps * (time_steps - 1) // 2)
        + number_of_facilities * number_of_tasks
        + number_of_facilities * number_of_tasks * time_steps
    )
    assert constraints_lowerbound <= model.getNConss() <= constraints_upperbound
    assert model.getObjectiveSense() == "minimize"

    model.optimize()


@pytest.mark.parametrize(
    "n_resources,n_tasks,seed1,seed2",
    itertools.product([1, 2, 3], [5, 10, 15], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_param_generation_seeding(n_resources, n_tasks, seed1, seed2):
    params1 = generate_params(n_resources, n_tasks, seed=seed1)
    params2 = generate_params(n_resources, n_tasks, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


def test_hooker_simple_instance():
    hooker_model = late_tasks_formulation(*_simple_instance_params()[:-1])
    hooker_model.hideOutput()
    hooker_model.optimize()
    assert hooker_model.getStatus() == "optimal"
    assert hooker_model.getObjVal() == 0


def test_heinz_simple_instance():
    params = (
        n_resources,
        n_tasks,
        time_steps,
        processing_times,
        capacities,
        assignment_costs,
        release_times,
        deadlines,
        resource_requirements,
    ) = _simple_instance_params()
    heinz_model = heinz_formulation(
        n_resources,
        n_tasks,
        processing_times,
        capacities,
        assignment_costs,
        release_times,
        deadlines,
        resource_requirements,
    )
    heinz_model.hideOutput()
    heinz_model.optimize()
    assert heinz_model.getStatus() == "optimal"
    assert heinz_model.getObjVal() == 1


def _simple_instance_params():
    n_resources = 1
    n_tasks = 1
    time_steps = 1
    processing_times = {(0, 0): 1}
    capacities = [1]
    assignment_costs = {(0, 0): 1}
    release_times = [0]
    # change the deadline to 0 or 1 to make it infeasible for the heinz formulation
    # and add 1 late task to the hooker formulation
    deadlines = [2]
    resource_requirements = {(0, 0): 1}
    return (
        n_resources,
        n_tasks,
        time_steps,
        processing_times,
        capacities,
        assignment_costs,
        release_times,
        deadlines,
        resource_requirements,
    )


def _check_hookers_instance(model, number_of_facilities, number_of_tasks, time_steps):
    assert (
        model.getNVars()
        == number_of_facilities * number_of_tasks * time_steps + number_of_tasks
    )
    constraints_lowerbound = (
        number_of_tasks * time_steps
        + number_of_tasks
        + number_of_facilities * time_steps
    )
    constraints_upperbound = (
        number_of_tasks * time_steps
        + number_of_tasks
        + number_of_facilities * time_steps
        + number_of_facilities * number_of_tasks * time_steps
    )
    assert constraints_lowerbound <= model.getNConss() <= constraints_upperbound
    assert model.getObjectiveSense() == "minimize"


def check_params_dimensions(params):
    (
        number_of_facilities,
        number_of_tasks,
        processing_times,
        capacities,
        assignment_costs,
        release_times,
        deadlines,
        resource_requirements,
    ) = params
    facility_for_task_count = number_of_facilities * number_of_tasks
    assert len(processing_times) == facility_for_task_count
    assert len(assignment_costs) == facility_for_task_count
    assert len(resource_requirements) == facility_for_task_count
    assert len(release_times) == number_of_tasks
    assert len(deadlines) == number_of_tasks
    assert len(capacities) == number_of_facilities


def check_params_ranges(params, params_ranges):
    for param, (start, end) in zip(params, params_ranges):
        if isinstance(param, int):
            assert start <= param < end
        elif isinstance(param, dict):
            for val in param.values():
                assert start <= val < end


def test_c_params_generation():
    n = 0
    for params in c_instance_params():
        n += 1
        check_params_dimensions(params)
        check_params_ranges(
            params,
            [
                (2, 4 + 1),
                (10, 38 + 1),
                (1, 10 * 4),
                (10, 10 + 1),
                (2, 20 * 4),
                (0, 0 + 1),
                (21, 95 + 1),
                (1, 10),
            ],
        )
    assert n == 3 * 15


def test_e_params_generation():
    n = 0
    for params in e_instance_params():
        n += 1
        check_params_dimensions(params)
        check_params_ranges(
            params,
            [
                (2, 10 + 1),
                (10, 5 * 10 + 1),
                (2, 25),
                (10, 10 + 1),
                (16, 53 + 1),
                (0, 0 + 1),
                (33, 33 + 1),
                (1, 10),
            ],
        )
    assert n == 9


def test_de_params_generation():
    n = 0
    for params in de_instance_params():
        n += 1
        check_params_dimensions(params)
        check_params_ranges(
            params,
            [
                (3, 3 + 1),
                (14, 28 + 1),
                (2, 30),
                (10, 10 + 1),
                (10, 60),
                (0, 0 + 1),
                (6, 95 + 1),
                (1, 10),
            ],
        )
    assert n == 8


def test_df_params_generation():
    n = 0
    for params in df_instance_params():
        n += 1
        check_params_dimensions(params)
        check_params_ranges(
            params,
            [
                (3, 3 + 1),
                (14, 28 + 1),
                (2, 30),
                (10, 10 + 1),
                (10, 60),
                (0, 4 + 1),
                (2, math.inf),
                (1, 10),
            ],
        )
    assert n == 8


@pytest.mark.parametrize(
    "number_of_facilities,number_of_tasks", itertools.product([2, 3, 4], [10, 20, 30])
)
def test_hooker_cost_formulation(number_of_facilities, number_of_tasks):
    params = c_params_generator(number_of_facilities, number_of_tasks)
    model = hooker_cost_formulation(*params)
    time_steps = range(min(params[5]), int(max(params[6])))
    constraints_lower_bound = number_of_tasks + number_of_facilities * len(time_steps)
    constraints_upper_bound = (
        constraints_lower_bound
        + number_of_facilities * number_of_tasks * len(time_steps)
    )
    assert model.getNVars() == number_of_facilities * number_of_tasks * len(time_steps)
    assert constraints_lower_bound <= model.getNConss() <= constraints_upper_bound



import geco.mips.facility_location as facility_location
import geco.mips.scheduling as scheduling
import geco.mips.set_packing as set_packing
import geco.mips.set_cover as set_cover
import geco.mips.independent_set as independent_set
import geco.mips.knapsack as knapsack
import geco.mips.max_cut as max_cut
import geco.mips.production_planning as production_planning
import geco.mips.packing as packing
import geco.mips.graph_coloring as graph_coloring
import geco.mips.loading as miplib
import geco.mips.combinatorial_auction as combinatorial_auction



import pyscipopt as scip
from networkx.utils import py_random_state
from geco.mips.packing.generic import *


@py_random_state(-1)
def tang_instance(n, m, binary=False, seed=0):
    """Generates a packing instance as described in A.2 in [1].

    Parameters:
    ----------
    n: int
        number of variables
    m: int
        number of constraints
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References:
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    return packing(n, m, *tang_params(n, m, binary, seed), binary, name="Tang Packing")


@py_random_state(-1)
def tang_params(n, m, binary, seed=0):
    """Generates a packing instance as described in A.2 in [1].

    Parameters:
    ----------
    n: int
        Number of variables
    m: int
        Number of constraints
    binary: bool
        Use binary variables coefficients or (non-negative) integer variables coefficients
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    costs: list[number] of size n
        Coefficients of objective function
    constraint_coefficients: list[list[number]] of dimensions (m x n)
        Coefficients of each variable for each constraint
    limits: list[number] of size m
        Limits of each constraint

    References:
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    costs = [seed.randint(1, 10) for _ in range(n)]

    if binary:
        constraint_coefficients = [
            [seed.randint(5, 30) for _ in range(n)] for _ in range(m)
        ]
        limits = [seed.randint(10 * n, 20 * n) for _ in range(m)]
    else:
        constraint_coefficients = [
            [seed.randint(0, 5) for _ in range(n)] for _ in range(m)
        ]
        limits = [seed.randint(9 * n, 10 * n) for _ in range(m)]

    return costs, constraint_coefficients, limits



import numpy as np
from networkx.utils import np_random_state

from geco.mips.combinatorial_auction.generic import combinatorial_auction


@np_random_state(-1)
def gasse_instance(
    n_items=100,
    n_bids=500,
    min_value=1,
    max_value=100,
    value_deviation=0.5,
    add_item_prob=0.9,
    max_n_sub_bids=5,
    additivity=0.2,
    budget_factor=1.5,
    resale_factor=0.5,
    integers=False,
    warnings=False,
    seed=0,
):
    """
    Generate a Combinatorial Auction instance following the 'arbitrary' scheme found in section 4.3. of [1].

    Parameters
    ----------
    n_items: int
        The number of items
    n_bids: int
        The number of bids
    min_value: int
        The minimum resale value for an item
    max_value: int
        The maximum resale value for an item
    value_deviation: float
        The deviation allowed for each bidder's private value of an item, relative from max_value
    add_item_prob: float between 0 and 1
        The probability of adding a new item to an existing bundle
    max_n_sub_bids: int
        The maximum number of substitutable bids per bidder (+1 gives the maximum number of bids per bidder)
    additivity: float
        Additivity parameter for bundle prices. Note that additivity < 0 gives sub-additive bids, while additivity > 0 gives super-additive bids
    budget_factor: float
        The budget factor for each bidder, relative to their initial bid's price
    resale_factor: float
        The resale factor for each bidder, relative to their initial bid's resale value
    integers: logical
        Should bid's prices be integral ?
    warnings: logical
        Should warnings be printed ?
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Kevin Leyton-Brown, Mark Pearson, and Yoav Shoham. (2000).
    Towards a universal test suite for combinatorial auction algorithms.
    Proceedings of ACM Conference on Electronic Commerce (EC-00) 66-76.
    """
    return combinatorial_auction(
        *gasse_params(
            n_items,
            n_bids,
            min_value,
            max_value,
            value_deviation,
            add_item_prob,
            max_n_sub_bids,
            additivity,
            budget_factor,
            resale_factor,
            integers,
            warnings,
            seed,
        ),
        n_items=n_items,
        name="Gasse Combinatorial Auction"
    )


@np_random_state(-1)
def gasse_params(
    n_items=100,
    n_bids=500,
    min_value=1,
    max_value=100,
    value_deviation=0.5,
    add_item_prob=0.9,
    max_n_sub_bids=5,
    additivity=0.2,
    budget_factor=1.5,
    resale_factor=0.5,
    integers=False,
    warnings=False,
    seed=0,
):
    """
    Generate a Combinatorial Auction instance params following the 'arbitrary' scheme found in section 4.3. of [1].
    This is based on the code from [2].

    Parameters
    ----------
    n_items: int
        The number of items
    n_bids: int
        The number of bids
    min_value: int
        The minimum resale value for an item
    max_value: int
        The maximum resale value for an item
    value_deviation: float
        The deviation allowed for each bidder's private value of an item, relative from max_value
    add_item_prob: float between 0 and 1
        The probability of adding a new item to an existing bundle
    max_n_sub_bids: int
        The maximum number of substitutable bids per bidder (+1 gives the maximum number of bids per bidder)
    additivity: float
        Additivity parameter for bundle prices. Note that additivity < 0 gives sub-additive bids, while additivity > 0 gives super-additive bids
    budget_factor: float
        The budget factor for each bidder, relative to their initial bid's price
    resale_factor: float
        The resale factor for each bidder, relative to their initial bid's resale value
    integers: logical
        Should bid's prices be integral ?
    warnings: logical
        Should warnings be printed ?
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    bids: list[tuple[list, float]]
        A list of bids each represented by a tuple of a bundle and the price proposed
    n_dummy_items: int
        Number of dummy items added to each bid

    References
    ----------
    .. [1] Kevin Leyton-Brown, Mark Pearson, and Yoav Shoham. (2000).
    Towards a universal test suite for combinatorial auction algorithms.
    Proceedings of ACM Conference on Electronic Commerce (EC-00) 66-76.
    .. [2] https://github.com/ds4dm/learn2branch/blob/master/01_generate_instances.py
    """
    assert min_value >= 0 and max_value >= min_value
    assert add_item_prob >= 0 and add_item_prob <= 1

    def choose_next_item(bundle_mask, interests, compats, add_item_prob, seed):
        n_items = len(interests)
        prob = (1 - bundle_mask) * interests * compats[bundle_mask, :].mean(axis=0)
        prob /= prob.sum()
        return seed.choice(n_items, p=prob)

    # common item values (resale price)
    values = min_value + (max_value - min_value) * seed.rand(n_items)

    # item compatibilities
    compats = np.triu(seed.rand(n_items, n_items), k=1)
    compats = compats + compats.transpose()
    compats = compats / compats.sum(1)

    bids = []
    n_dummy_items = 0

    # create bids, one bidder at a time
    while len(bids) < n_bids:

        # bidder item values (buy price) and interests
        private_interests = seed.rand(n_items)
        private_values = values + max_value * value_deviation * (
            2 * private_interests - 1
        )

        # substitutable bids of this bidder
        bidder_bids = {}

        # generate initial bundle, choose first item according to bidder interests
        prob = private_interests / private_interests.sum()
        item = seed.choice(n_items, p=prob)
        bundle_mask = np.full(n_items, 0)
        bundle_mask[item] = 1

        # add additional items, according to bidder interests and item compatibilities
        while seed.rand() < add_item_prob:
            # stop when bundle full (no item left)
            if bundle_mask.sum() == n_items:
                break
            item = choose_next_item(
                bundle_mask, private_interests, compats, add_item_prob, seed
            )
            bundle_mask[item] = 1

        bundle = np.nonzero(bundle_mask)[0]

        # compute bundle price with value additivity
        price = private_values[bundle].sum() + np.power(len(bundle), 1 + additivity)
        if integers:
            price = int(price)

        # drop negativaly priced bundles
        if price < 0:
            if warnings:
                print("warning: negatively priced bundle avoided")
            continue

        # bid on initial bundle
        bidder_bids[frozenset(bundle)] = price

        # generate candidates substitutable bundles
        sub_candidates = []
        for item in bundle:

            # at least one item must be shared with initial bundle
            bundle_mask = np.full(n_items, 0)
            bundle_mask[item] = 1

            # add additional items, according to bidder interests and item compatibilities
            while bundle_mask.sum() < len(bundle):
                item = choose_next_item(
                    bundle_mask, private_interests, compats, add_item_prob, seed
                )
                bundle_mask[item] = 1

            sub_bundle = np.nonzero(bundle_mask)[0]

            # compute bundle price with value additivity
            sub_price = private_values[sub_bundle].sum() + np.power(
                len(sub_bundle), 1 + additivity
            )
            if integers:
                sub_price = int(sub_price)

            sub_candidates.append((sub_bundle, sub_price))

        # filter valid candidates, higher priced candidates first
        budget = budget_factor * price
        min_resale_value = resale_factor * values[bundle].sum()
        for bundle, price in [
            sub_candidates[i]
            for i in np.argsort([-price for bundle, price in sub_candidates])
        ]:

            if (
                len(bidder_bids) >= max_n_sub_bids + 1
                or len(bids) + len(bidder_bids) >= n_bids
            ):
                break

            if price < 0:
                if warnings:
                    print("warning: negatively priced substitutable bundle avoided")
                continue

            if price > budget:
                if warnings:
                    print("warning: over priced substitutable bundle avoided")
                continue

            if values[bundle].sum() < min_resale_value:
                if warnings:
                    print(
                        "warning: substitutable bundle below min resale value avoided"
                    )
                continue

            if frozenset(bundle) in bidder_bids:
                if warnings:
                    print("warning: duplicated substitutable bundle avoided")
                continue

            bidder_bids[frozenset(bundle)] = price

        # add XOR constraint if needed (dummy item)
        if len(bidder_bids) > 2:
            dummy_item = [n_items + n_dummy_items]
            n_dummy_items += 1
        else:
            dummy_item = []

        # place bids
        for bundle, price in bidder_bids.items():
            bids.append((list(bundle) + dummy_item, price))

        return bids, n_dummy_items



import itertools
import tempfile

import pyscipopt as scip


def shuffle(model, seed, cons=True, vars=True):
    """
    Shuffles a MIP instance's rows & columns

    Parameters
    ----------
    model: scip.Model
        A pyscipopt model of the to be shuffled instance
    seed: int
        Used in shuffling (must be bigger than 0)
    cons: bool
        Whether the columns should be shuffled
    vars: bool
        Whether the rows should be shuffled

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the shuffled instance
    """
    # The following line of code does not correctly set the name! Leave commented until it's clear why.
    # shuffled = scip.Model(sourceModel=model, problemName=model.getProbName(), origcopy=True)
    assert seed > 0
    shuffled = scip.Model()
    with tempfile.NamedTemporaryFile(suffix=".mps") as temp:
        model.writeProblem(temp.name)
        shuffled.setParam("randomization/permutationseed", seed)
        shuffled.setParam("randomization/permuteconss", cons)
        shuffled.setParam("randomization/permutevars", vars)
        shuffled.readProblem(temp.name)
        shuffled.setProbName(model.getProbName())
    return shuffled


def expand_parameters(function, **parameter_lists):
    """
    Calls a function with every combination of params

    Parameters
    ----------
    function: function
    parameter_lists: dict[str,list]
        Maps parameter name to all values it might take

    Returns
    -------
    generator: Generator
        Generator of returned values from function with each parameter combination
    """
    parameter_names = parameter_lists.keys()
    all_possible_parameters = itertools.product(*parameter_lists.values())
    for params in all_possible_parameters:
        yield function(**{name: val for name, val in zip(parameter_names, params)})



import itertools

import pytest

from geco.mips.independent_set.barabasi_albert import *
from geco.mips.independent_set.gasse import *
from geco.mips.independent_set.generic import *


@pytest.mark.parametrize(
    "graph",
    [
        nx.generators.complete_graph(3),
        nx.generators.complete_graph(10),
        nx.generators.complete_graph(50),
    ],
)
def test_independent_set(graph):
    model = independent_set(graph)
    n = len(graph.nodes)
    assert model.getNVars() == n
    assert model.getNConss() == n * (n - 1) / 2
    assert model.getObjectiveSense() == "maximize"


def test_simple_instance():
    graph = nx.generators.complete_graph(3)
    model = independent_set(graph)
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 1


@pytest.mark.parametrize(
    "graph",
    [
        nx.generators.complete_graph(3),
        nx.generators.complete_graph(10),
        nx.generators.complete_graph(50),
    ],
)
def test_clique_independent_set(graph):
    model = clique_independent_set(graph)
    assert model.getNVars() == len(graph.nodes)
    assert model.getNConss() <= len(graph.edges)
    assert model.getObjectiveSense() == "maximize"


def test_clique_simple_instance():
    graph = nx.generators.complete_graph(3)
    model = clique_independent_set(graph)
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 1


@pytest.mark.parametrize(
    "n,p,seed", itertools.product([10, 50], [0.5, 0.9], [0, 1, 1337, 53115])
)
def test_gasse(n, p, seed):
    model = gasse_instance(n, p, seed)
    assert model.getNVars() == n
    assert model.getNConss() <= n * (n - 1) / 2
    assert model.getObjectiveSense() == "maximize"


@pytest.mark.parametrize(
    "n,p,seed1,seed2",
    itertools.product([10, 50], [0.5, 0.9], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_gasse_seeding(n, p, seed1, seed2):
    graph1 = gasse_params(n, p, seed=seed1)
    graph2 = gasse_params(n, p, seed=seed2)
    same_edges = set(graph1.edges) == set(graph2.edges)
    same_seeds_produce_same_params = seed1 == seed2 and same_edges
    different_seeds_produce_different_params = seed1 != seed2 and not same_edges
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


@pytest.mark.parametrize(
    "n,m,seed", itertools.product([10, 50], [5, 9], [0, 1, 1337, 53115])
)
def test_barabasi_albert(n, m, seed):
    model = barabasi_albert_instance(n, m, seed)
    assert model.getNVars() == n
    assert model.getNConss() <= n * (n - 1) / 2
    assert model.getObjectiveSense() == "maximize"


@pytest.mark.parametrize(
    "n,m,seed1,seed2",
    itertools.product([100, 500], [5, 10], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_barabasi_albert_seeding(n, m, seed1, seed2):
    graph1 = barabasi_albert_params(n, m, seed=seed1)
    graph2 = barabasi_albert_params(n, m, seed=seed2)
    same_edges = set(graph1.edges) == set(graph2.edges)
    same_degrees = dict(graph1.degree) == dict(graph2.degree)
    same = same_edges and same_degrees
    same_seeds_produce_same_params = seed1 == seed2 and same
    different_seeds_produce_different_params = seed1 != seed2 and not same
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



import pyscipopt as scip


def uncapacitated_lot_sizing(
    T, M, initial_storage, final_storage, p, h, q, d, name="Production Planning"
):
    """
    Generates an uncapacitated lot-sizing MIP instance instance as in 2.1 of [1].

    Parameters
    ----------
    T: int
        Time horizon
    M: int
        Maximum lot size at any time step
    initial_storage: int
        Initial available storage
    final_storage: int
        Storage available at the last time step
    p: list[int]
        Unit production cost at each time step
    h: list[int]
        Unit inventory cost at each time step
    q: list[int]
        Fixed production cost at each time step
    d: list[int]
        Demand at each time step
    name: str
        Name to be given to the generated model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Pochet, Y. and Wolsey, L. A. (2006). Production planning by
    mixed integer programming. Springer Science & Business Media.
    """
    model = scip.Model(name)
    # add variables and their cost
    production_vars = []
    produce_or_not_vars = []
    storage_vars = []
    for i in range(T + 1):
        var = model.addVar(lb=0, ub=None, obj=p[i], name=f"x_{i}", vtype="I")
        production_vars.append(var)

        var = model.addVar(lb=0, ub=1, obj=h[i], name=f"y_{i}", vtype="B")
        produce_or_not_vars.append(var)

        var = model.addVar(lb=0, ub=None, obj=q[i], name=f"s_{i}", vtype="I")
        storage_vars.append(var)

    # remove unneeded var
    model.delVar(production_vars[0])

    # add constraints
    for i in range(1, T + 1):
        model.addCons(
            storage_vars[i - 1] + production_vars[i] == d[i] + storage_vars[i]
        )
        model.addCons(production_vars[i] <= M * produce_or_not_vars[i])

    model.addCons(storage_vars[0] == initial_storage)
    model.addCons(storage_vars[T] == final_storage)

    model.setMinimize()

    return model



import itertools

import pytest

from geco.graphs.chimera import *


@pytest.mark.parametrize(
    "m, n, t",
    itertools.product(
        [2, 4, 6],
        [2, 4, 6],
        [2, 4, 6],
    ),
)
def test_chimera(m, n, t):
    g = chimera_graph(n, m, t)
    assert g.number_of_nodes() == m * n * t * 2
    assert g.number_of_edges() == n * m * t * t + n * (m - 1) * t + m * (n - 1) * t


@pytest.mark.parametrize(
    "m, n, t",
    itertools.product(
        [2, 4, 6],
        [2, 4, 6],
        [2, 4, 6],
    ),
)
def test_dwave_chimera(m, n, t):
    g = dwave_chimera_graph(m, n, t)
    assert g.number_of_nodes() == m * n * t * 2
    assert g.number_of_edges() == n * m * t * t + n * (m - 1) * t + m * (n - 1) * t


@pytest.mark.parametrize("m", [3, 5, 10])
def test_selby(m):
    graph = selby_c(m)
    assert graph.number_of_nodes() == m * m * 8
    assert graph.number_of_edges() == 24 * m * m - 8 * m


@pytest.mark.parametrize("m, faulty", itertools.product([3, 8, 10], [20, 50, 73]))
def test_mgw(m, faulty):
    graph = mgw(m=m, faulty=faulty)
    assert graph.number_of_nodes() == m * m * 8
    max_degree = 6
    maximum_edges_connected_to_faulty_nodes = faulty * max_degree
    assert (
        graph.number_of_edges()
        >= (24 * m * m - 8 * m) - maximum_edges_connected_to_faulty_nodes
    )



from geco.graphs.pegasus import *

import pytest


@pytest.fixture
def g_dwave():
    return dwave_pegasus_graph(6)


def test_nodes(g_dwave):
    # calculated from 8(3M − 1)(M − 1)
    assert g_dwave.number_of_nodes() == 8 * (3 * 6 - 1) * (6 - 1)


def test_edges(g_dwave):
    # from the p6 example in dwave's paper
    assert g_dwave.number_of_edges() == 4484



import networkx as nx
import pytest

from geco.mips.graph_coloring.generic import *


def _test_cycle_instance_output(model):
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 3


def test_assignment():
    graph = nx.cycle_graph(5)
    H = 5
    model = assignment(graph, H)
    n = len(graph.nodes)
    m = len(graph.edges)

    # test variables and constraints
    assert model.getNVars() == n * H + H
    assert model.getNConss() == n + m * H

    _test_cycle_instance_output(model)


def test_assignment_asymmetric():
    graph = nx.cycle_graph(5)
    H = 5
    model = assignment_asymmetric(graph, H)
    n = len(graph.nodes)
    m = len(graph.edges)

    # test variables and constraints
    assert model.getNVars() == n * H + H
    assert model.getNConss() == n + m * H + H + (H - 1)

    _test_cycle_instance_output(model)


def test_representatives():
    graph = nx.cycle_graph(5)
    model = representatives(graph)
    n = len(graph.nodes)
    m = len(graph.edges)

    # test variables and constraints
    assert n <= model.getNVars() <= n ** 2
    assert n <= model.getNConss() <= n + n * m

    _test_cycle_instance_output(model)


def test_set_covering():
    graph = nx.cycle_graph(5)
    subsets = [{0, 1}, {2, 3}, {4}]
    model = set_covering(graph, subsets)
    n = len(graph.nodes)

    # test variables and constraints
    assert 1 <= model.getNVars() <= 2 ** n
    assert n <= model.getNConss() <= n * (2 ** n)

    _test_cycle_instance_output(model)


def test_partial_ordering():
    graph = nx.cycle_graph(5)
    H = 5
    model = partial_ordering(graph, H)
    n = len(graph.nodes)

    # test variables and constraints
    assert model.getNVars() == 2 * n * H
    assert n <= model.getNConss() == 2 * n + 4 * n * (H - 1)

    _test_cycle_instance_output(model)


def test_hybrid_partial_ordering():
    graph = nx.cycle_graph(5)
    H = 5
    model = hybrid_partial_ordering(graph, H)
    n = len(graph.nodes)

    # test variables and constraints
    assert model.getNVars() == 3 * n * H
    assert n + n * H <= model.getNConss() == 2 * n + 3 * n * (H - 1) + n * H + n * H

    _test_cycle_instance_output(model)


def test_node_labeling():
    g = nx.karate_club_graph()
    g = nx.convert_node_labels_to_integers(g, first_label=1)
    H = 10
    models = [
        representatives(g),
        assignment(g, H),
        assignment_asymmetric(g, H),
        partial_ordering(g, H),
        hybrid_partial_ordering(g, H),
    ]
    for m in models:
        m.hideOutput()
        m.optimize()
        assert m.getStatus() == "optimal" and m.getObjVal() == 5



import collections
import itertools

import pytest

from geco.mips.set_cover.yang import *
from geco.mips.set_cover.sun import *
from geco.mips.set_cover.orlib import *
from geco.mips.set_cover.gasse import *

"""
Generic Tests
"""


def test_set_cover_solution_1():
    model = set_cover([1], [{0}])
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 1


def test_set_cover_solution_2():
    model = set_cover([1, 1, 1], [{0}, {1}, {2}])
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 3


"""
Yang Tests
"""


@pytest.mark.parametrize(
    "m,seed", itertools.product([10, 100, 200], [0, 1, 1337, 53115])
)
def test_yang_set_cover_creation(m, seed):
    model = yang_instance(m, seed)
    assert model.getNVars() == 10 * m
    assert model.getNConss() == m
    assert model.getObjectiveSense() == "minimize"


@pytest.mark.parametrize(
    "m,seed1,seed2",
    itertools.product([10, 100, 200], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_yang_parameter(m, seed1, seed2):
    params1 = yang_params(m, seed=seed1)
    params2 = yang_params(m, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


"""
Sun Tests
"""


@pytest.mark.parametrize(
    "n,m,seed", itertools.product([10, 100, 200], [10, 100, 200], [0, 1, 1337, 53115])
)
def test_sun_set_cover_creation(n, m, seed):
    model = sun_instance(n, m, seed)
    assert model.getNVars() == n
    assert model.getNConss() == m
    assert model.getObjectiveSense() == "minimize"


@pytest.mark.parametrize(
    "n,m,seed1,seed2",
    itertools.product(
        [10, 100, 200], [10, 100, 200], [0, 1, 1337, 53115], [0, 1, 1337, 53115]
    ),
)
def test_sun_params(n, m, seed1, seed2):
    params1 = sun_params(n, m, seed=seed1)
    params2 = sun_params(n, m, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


@pytest.mark.parametrize(
    "n,m,seed", itertools.product([10, 100, 200], [10, 100, 200], [0, 1, 1337, 53115])
)
def test_sun_at_least_two_elements_in_set(n, m, seed):
    _, sets = sun_params(n, m, seed=seed)
    counter = collections.defaultdict(int)
    for s in sets:
        for e in s:
            counter[e] += 1
    assert all([count >= 2 for count in counter.values()])


@pytest.mark.parametrize(
    "n,base_n,base_m,seed1,seed2",
    itertools.product(
        [10, 100, 200],
        [1, 5, 9],
        [10, 100, 200],
        [0, 1, 1337, 53115],
        [0, 1, 1337, 53115],
    ),
)
def test_expand_sun_params(n, base_n, base_m, seed1, seed2):
    base_costs1, base_sets1 = sun_params(base_n, base_m, seed1)
    base_costs2, base_sets2 = sun_params(base_n, base_m, seed2)
    params1 = costs1, sets1 = expand_sun_params((n,), (base_costs1, base_sets1), seed1)
    params2 = costs2, sets2 = expand_sun_params((n,), (base_costs2, base_sets2), seed2)

    # test seeding
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params

    # test correct size
    assert len(costs1) == len(costs2) == n
    assert len(sets1) == len(sets2) == base_m


"""
OR-Library tests
"""


def test_scp_orlib():
    instance_name = "scp41.txt"
    instance = orlib_instance(instance_name)
    assert instance.getNVars() == 1000
    assert instance.getNConss() == 200


def test_rail_orlib():
    instance_name = "rail507.txt"
    instance = orlib_instance(instance_name)
    assert instance.getNVars() == 63009
    assert instance.getNConss() == 507


"""
Gasse tests
"""


@pytest.mark.parametrize(
    "nrows,ncols,density,seed1,seed2",
    itertools.product(
        [100, 200],
        [10, 100, 200],
        [0.2, 0.3, 0.5],
        [0, 1, 1337, 53115],
        [0, 1, 1337, 53115],
    ),
)
def test_gasse_params(nrows, ncols, density, seed1, seed2):
    params1 = gasse_params(nrows, ncols, density, seed=seed1)
    params2 = gasse_params(nrows, ncols, density, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


@pytest.mark.parametrize(
    "nrows,ncols,density,seed",
    itertools.product(
        [100, 200],
        [50, 70],
        [0.2, 0.3, 0.5],
        [0, 1, 1337, 53115],
    ),
)
def test_gasse_instance(nrows, ncols, density, seed):
    model = gasse_instance(nrows, ncols, density, max_coef=10, seed=seed)
    assert model.getNVars() == ncols
    assert model.getNConss() == nrows
    assert model.getObjectiveSense() == "minimize"



import pyscipopt as scip
import itertools
import networkx as nx


def assignment(
    graph, color_upperbound, name="Assignment Graph Coloring", with_variables=False
):
    """
    Generates a graph coloring ILP formulation (ASS-S) as described in [1].

    Parameters
    ----------
    graph: networkx graph
        Input graph
    color_upperbound: int
        Maximum number of colors to use
    name: str
        Name of the model
    with_variables: bool
        return variables with the generated model (used to extend model)

    Returns
    -------
       model: scip.Model
        pyscipopt model of the generated instance

    References
    ----------
    .. [1] A.Jabrayilov, P.Mutzel "New Integer Linear Programming Models for the Vertex Coloring Problem"
    """
    model = scip.Model(name)

    graph = nx.convert_node_labels_to_integers(graph, first_label=0)
    colors = range(color_upperbound)

    # add variables and their cost
    x = {
        (vertex, color): model.addVar(
            lb=0, ub=1, obj=0, name=f"x_{vertex}_{color}", vtype="B"
        )
        for vertex, color in itertools.product(graph.nodes, colors)
    }

    w = {
        color: model.addVar(lb=0, ub=1, obj=1, name=f"w_{color}", vtype="B")
        for color in colors
    }

    # add constraint (2)
    for v in graph.nodes:
        model.addCons(scip.quicksum(x[v, color] for color in colors) == 1)

    # add constraint (3)
    for u, v in graph.edges:
        for color in colors:
            model.addCons(x[u, color] + x[v, color] <= w[color])

    model.setMinimize()
    if with_variables:
        return model, w, x
    else:
        return model


def assignment_asymmetric(
    graph, color_upperbound, name="Assignment Extended Graph Coloring"
):
    """
    Generates a graph coloring ILP formulation (ASS) as described in [1] which
    uses symmetry breaking rules from [2].

    Parameters
    ----------
    graph: networkx graph
        Input graph
    color_upperbound: int
        Maximum number of colors to use
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] A.Jabrayilov, P.Mutzel "New Integer Linear Programming Models for the Vertex Coloring Problem"
    .. [2] I.Mendez, P.Zabala "A Branch-and-Cut Algorithm for Graph Coloring"
    """
    model, w, x = assignment(graph, color_upperbound, name, with_variables=True)

    graph = nx.convert_node_labels_to_integers(graph, first_label=0)
    colors = list(range(color_upperbound))

    # add constraint (5)
    for color in colors:
        model.addCons(w[color] <= scip.quicksum(x[v, color] for v in graph.nodes))

    # add constraint (6)
    for color in colors[1:]:
        model.addCons(w[color] <= w[color - 1])
    return model


def representatives(graph, name="Representatives Graph Coloring"):
    """
    Generates a graph coloring ILP formulation (REP) as described in [1].

    Parameters
    ----------
    graph: networkx graph
        Input graph
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] A.Jabrayilov, P.Mutzel "New Integer Linear Programming Models for the Vertex Coloring Problem"
    """
    model = scip.Model(name)

    graph = nx.convert_node_labels_to_integers(graph, first_label=0)

    # add variables and their cost
    x = {}
    for (u, v) in itertools.product(graph.nodes, graph.nodes):
        if not graph.has_edge(u, v) or u == v:
            obj = 1 if u == v else 0
            x[u, v] = model.addVar(lb=0, ub=1, obj=obj, name=f"x_{u}_{v}", vtype="B")

    # add constraint (8)
    for v in graph.nodes:
        non_adjacent_vertices = (graph.nodes - graph.neighbors(v)).union({v})
        model.addCons(scip.quicksum(x[u, v] for u in non_adjacent_vertices) >= 1)

    # add constraint (9)
    for u in graph.nodes:
        non_adjacent_vertices = graph.nodes - graph.neighbors(u) - {u}
        for v, w in graph.edges:
            if v in non_adjacent_vertices and w in non_adjacent_vertices:
                model.addCons(x[u, v] + x[u, w] <= x[u, u])

    model.setMinimize()
    return model


def set_covering(graph, subsets, name="Set Covering Graph Coloring"):
    """
    Generates a graph coloring ILP formulation (COV) as described in [1].

    Parameters
    ----------
    graph: networkx graph
        Input graph
    subsets: Iterable of sets
        Independent Sets of nodes
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] A.Jabrayilov, P.Mutzel "New Integer Linear Programming Models for the Vertex Coloring Problem"
    """
    model = scip.Model(name)

    graph = nx.convert_node_labels_to_integers(graph, first_label=0)

    x = {
        tuple(s): model.addVar(lb=0, ub=1, obj=1, name=f"x_{s}", vtype="B")
        for s in subsets
    }

    for v in graph.nodes:
        model.addCons(scip.quicksum(x[tuple(s)] for s in subsets if v in s) >= 1)

    model.setMinimize()

    return model


def _partial_ordering_base_model(graph, colors, model):
    q = 0

    # add variables and their cost
    y = {
        (color, vertex): model.addVar(
            lb=0,
            ub=1,
            obj=1 if vertex == q else 0,
            name=f"y_{color}_{vertex}",
            vtype="B",
        )
        for vertex, color in itertools.product(graph.nodes, colors)
    }

    z = {
        (vertex, color): model.addVar(
            lb=0, ub=1, obj=0, name=f"z_{vertex}_{color}", vtype="B"
        )
        for vertex, color in itertools.product(graph.nodes, colors)
    }

    # add constraint (16) and (17)
    for v in graph.nodes:
        model.addCons(z[v, 1] == 0)
        model.addCons(y[colors[-1], v] == 0)

    # add constraint (18), (19) and (21)
    for v, c in itertools.product(graph.nodes, colors[:-1]):
        model.addCons(y[c, v] - y[c + 1, v] >= 0)
        model.addCons(y[c, v] + z[v, c + 1] == 1)
        model.addCons(y[c, q] - y[c, v] >= 0)

    model.setMinimize()

    return model, y, z


def partial_ordering(graph, color_upperbound, name="Partial Ordering Graph Coloring"):
    """
    Generates a graph coloring ILP formulation (POP) as described in [1].

    Parameters
    ----------
    graph: networkx graph
        Input graph
    color_upperbound: int
        Maximum number of colors to use
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] A.Jabrayilov, P.Mutzel "New Integer Linear Programming Models for the Vertex Coloring Problem"
    """
    model = scip.Model(name)

    graph = nx.convert_node_labels_to_integers(graph, first_label=0)
    colors = list(range(color_upperbound))

    model, y, z = _partial_ordering_base_model(graph, colors, model)

    # add constraint (20)
    for (u, v), c in itertools.product(graph.edges, colors[:-1]):
        model.addCons(y[c, u] + z[u, c] + y[c, v] + z[v, c] >= 1)

    return model


def hybrid_partial_ordering(
    graph, color_upperbound, name="Hybrid Partial Ordering Graph Coloring"
):
    """
    Generates a graph coloring ILP formulation (POP2) as described in [1].

    Parameters
    ----------
    graph: networkx graph
        Input graph
    color_upperbound: int
        Maximum number of colors to use
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] A.Jabrayilov, P.Mutzel "New Integer Linear Programming Models for the Vertex Coloring Problem"
    """
    model = scip.Model(name)

    graph = nx.convert_node_labels_to_integers(graph, first_label=0)
    colors = list(range(color_upperbound))

    # add variables and their cost
    x = {
        (vertex, color): model.addVar(
            lb=0, ub=1, obj=0, name=f"x_{vertex}_{color}", vtype="B"
        )
        for vertex, color in itertools.product(graph.nodes, colors)
    }

    model, y, z = _partial_ordering_base_model(graph, colors, model)

    # add constraint (14)
    for v, c in itertools.product(graph.nodes, colors):
        model.addCons(x[v, c] == 1 - (y[c, v] + z[v, c]))

    # add constraint (23)
    for (u, v), c in itertools.product(graph.edges, colors):
        model.addCons(x[u, c] + x[v, c] <= 1)

    return model



import pyscipopt as scip
from networkx.utils import py_random_state
from geco.mips.production_planning.generic import *


@py_random_state(-1)
def tang_instance(T, seed=0):
    """Generates a production planning instance as described in A.2 in [1].

    Parameters
    ----------
    T: int
        Time horizon
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    return uncapacitated_lot_sizing(
        T, *tang_params(T, seed), name="Tang Production Planning"
    )


@py_random_state(-1)
def tang_params(T, seed=0):
    """Generates production planning instance params as described in A.2 in [1].

    Parameters
    ----------
    T: int
        Time horizon
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    M: int
        Maximum lot size at any time step
    initial_storage: int
        Initial available storage
    final_storage: int
        Storage available at the last time step
    p: list[int]
        Unit production cost at each time step
    h: list[int]
        Unit inventory cost at each time step
    q: list[int]
        Fixed production cost at each time step
    d: list[int]
        Demand at each time step

    References
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    initial_storage = 0
    final_storage = 20
    M = 100
    p = []
    h = []
    q = []
    d = []
    for i in range(T + 1):
        p.append(seed.randint(1, 10))
        h.append(seed.randint(1, 10))
        q.append(seed.randint(1, 10))
        d.append(seed.randint(1, 10))
    return M, initial_storage, final_storage, p, h, q, d



import numpy as np
from geco.mips.facility_location.generic import capacitated_facility_location
from networkx.utils import py_random_state


@py_random_state(-1)
def cornuejols_instance(n_customers, n_facilities, ratio, seed=0):
    """
    Generates a Capacitated Facility Location MIP formulation following [1].

    Parameters
    ----------
    n_customers: int
        Number of customers
    n_facilities: int
        Number of facilities
    ratio: float
        Capacity / demand ratio
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Cornuejols G, Sridharan R, Thizy J-M (1991)
        A Comparison of Heuristics and Relaxations for the Capacitated Plant Location Problem.
        European Journal of Operations Research 50:280-297.
    """
    return capacitated_facility_location(
        n_customers,
        n_facilities,
        *cornuejols_params(n_customers, n_facilities, ratio, seed),
    )


@py_random_state(-1)
def cornuejols_params(n_customers, n_facilities, ratio, seed=0):
    """
    Generates a Capacitated Facility Location instance params following [1].

    This code is heavily based on the code available in [1] which was used in [2] and
    the generation techniques in [3].

    Parameters
    ----------
    n_customers: int
        Number of customers
    n_facilities: int
        Number of facilities
    ratio: float
        Capacity / demand ratio
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    trans_costs: numpy array [float]
        Matrix of transportation costs from customer i to facility j [i,j]
    demands: numpy array [int]
        Demands of each customer
    fixed_costs: numpy array [int]
        Fixed costs of operating each facility
    capacities: numpy array [int]
        Capacities of each facility

     References
    ----------
    .. [1] https://github.com/ds4dm/learn2branch
    .. [2] "Exact Combinatorial Optimization with Graph Convolutional Neural Networks" (2019)
      Maxime Gasse, Didier Chételat, Nicola Ferroni, Laurent Charlin and Andrea Lodi
      Advances in Neural Information Processing Systems 32 (2019)
    .. [3] Cornuejols G, Sridharan R, Thizy J-M (1991)
        A Comparison of Heuristics and Relaxations for the Capacitated Plant Location Problem.
        European Journal of Operations Research 50:280-297.
    """
    # locations for customers
    c_x = np.array([seed.random() for _ in range(n_customers)])
    c_y = np.array([seed.random() for _ in range(n_customers)])

    # locations for facilities
    f_x = np.array([seed.random() for _ in range(n_facilities)])
    f_y = np.array([seed.random() for _ in range(n_facilities)])

    demands = np.array([seed.randint(5, 35 + 1) for _ in range(n_customers)])
    capacities = np.array([seed.randint(10, 160 + 1) for _ in range(n_facilities)])
    fixed_costs = np.array(
        [seed.randint(100, 110 + 1) for _ in range(n_facilities)]
    ) * np.sqrt(capacities) + np.array(
        [seed.randint(0, 90 + 1) for _ in range(n_facilities)]
    )
    fixed_costs = fixed_costs.astype(int)

    # adjust capacities according to ratio
    total_demand = demands.sum()
    total_capacity = capacities.sum()
    capacities = capacities * ratio * total_demand / total_capacity
    capacities = capacities.astype(int)

    # transportation cost
    trans_costs = (
        np.sqrt(
            (c_x.reshape((-1, 1)) - f_x.reshape((1, -1))) ** 2
            + (c_y.reshape((-1, 1)) - f_y.reshape((1, -1))) ** 2
        )
        * 10
        * demands.reshape((-1, 1))
    )
    return trans_costs, demands, fixed_costs, capacities



import pyscipopt as scip
import geco.mips.loading.orlib as orlib
from geco.mips.facility_location.generic import capacitated_warehouse_location


def cap_numeric_reader(file):
    """
    Reads cap(NUMBER) Capacitated Warehouse Location instance params mentioned in [1].

    Parameters
    ----------
    file: file-like object

    Returns
    -------
    n_customers: int
        The desired number of customers
    n_facilities: int
        The desired number of facilities
    transportation_cost: numpy array [float]
        Matrix of transportation costs from customer i to facility j [i,j]
    demands: numpy array [int]
        Demands of each customer
    fixed_costs: numpy array [int]
        Fixed costs of operating each facility
    capacities: numpy array [int]
        Capacities of each facility

    References
    ----------
    ..[1] http://people.brunel.ac.uk/~mastjjb/jeb/orlib/capinfo.html
    """
    num_of_warehouses, num_of_customers = orlib.read_numbers(file.readline())
    capacities = []
    fixed_costs = []
    demands = []
    allocation_cost_per_warehouse = {}

    for _ in range(num_of_warehouses):
        capacity, fixed_cost = orlib.read_numbers(file.readline())
        capacities.append(capacity)
        fixed_costs.append(fixed_cost)

    for i in range(num_of_customers):
        demand = orlib.read_number(file.readline())
        allocation_costs = orlib.read_multiline_numbers(file, num_of_warehouses)
        demands.append(demand)
        for j, cost in enumerate(allocation_costs):
            allocation_cost_per_warehouse[i, j] = cost

    return (
        num_of_customers,
        num_of_warehouses,
        allocation_cost_per_warehouse,
        demands,
        fixed_costs,
        capacities,
    )


def cap_alpha_reader(file, capacity):
    """
    Reads cap(LETTER) Capacitated Warehouse Location instance params mentioned in [1].

    Parameters
    ----------
    file: file-like object

    Returns
    -------
    n_customers: int
        The desired number of customers
    n_facilities: int
        The desired number of facilities
    transportation_cost: numpy array [float]
        Matrix of transportation costs from customer i to facility j [i,j]
    demands: numpy array [int]
        Demands of each customer
    fixed_costs: numpy array [int]
        Fixed costs of operating each facility
    capacities: numpy array [int]
        Capacities of each facility

    References
    ----------
    ..[1] http://people.brunel.ac.uk/~mastjjb/jeb/orlib/capinfo.html
    """
    num_of_warehouses, num_of_customers = orlib.read_numbers(file.readline())
    capacities = []
    fixed_costs = []
    demands = []
    allocation_cost_per_warehouse = {}

    for _ in range(num_of_warehouses):
        capacity, fixed_cost = capacity, float(file.readline().strip().split(b" ")[-1])
        capacities.append(capacity)
        fixed_costs.append(fixed_cost)

    for i in range(num_of_customers):
        demand = orlib.read_number(file.readline())
        allocation_costs = orlib.read_multiline_numbers(file, num_of_warehouses)
        demands.append(demand)
        for j, cost in enumerate(allocation_costs):
            allocation_cost_per_warehouse[i, j] = cost

    return (
        num_of_customers,
        num_of_warehouses,
        allocation_cost_per_warehouse,
        demands,
        fixed_costs,
        capacities,
    )


PSET_CAPACITIES = {
    "a": [8_000, 10_000, 12_000, 14_000],
    "b": [5_000, 6_000, 7_000, 8_000],
    "c": [5_000, 5_750, 6_500, 7_250],
}


def orlib_instance(instance_name):
    """
    Loads an orlib Capacitated Warehouse Location instance

    Parameters
    ----------
    instance_name: str
        Name of the capacitated warehouse location file. example: "cap41.txt"
        for files from problem sets A,B,C. To get the first problem of problem set A
        pass instance_name as "capa1.txt"

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the loaded instance
    """
    # TODO: assert that instance_name correlated to one of the listed capacitated warehouse location files
    if instance_name[:3] == "cap" and instance_name[3].isnumeric():
        return orlib.orlib_load_instance(
            instance_name,
            reader=cap_numeric_reader,
            formulation=capacitated_warehouse_location,
        )
    elif instance_name[:3] == "cap" and instance_name[3].isalpha():
        problem_number = int(instance_name[4])
        problem_set = instance_name[3]
        capacity = PSET_CAPACITIES[problem_set][problem_number - 1]
        return orlib.orlib_load_instance(
            instance_name[:4] + ".txt",
            reader=lambda file: cap_alpha_reader(file, capacity),
            formulation=capacitated_warehouse_location,
        )
    else:
        raise ValueError(
            f'"{instance_name}" is not a valid file name for Capacitated Warehouse Location'
            f"files listed in http://people.brunel.ac.uk/~mastjjb/jeb/orlib/capinfo.html"
        )



import itertools

import pyscipopt as scip


def capacitated_facility_location(
    n_customers,
    n_facilities,
    transportation_cost,
    demands,
    fixed_costs,
    capacities,
    name="Capacitated Facility Location",
):
    """
    Generate a Capacitated Facility Location MIP formulation following [1].

    Parameters
    ----------
    n_customers: int
        The desired number of customers
    n_facilities: int
        The desired number of facilities
    transportation_cost: numpy array [float]
        Matrix of transportation costs from customer i to facility j [i,j]
    demands: numpy array [int]
        Demands of each customer
    fixed_costs: numpy array [int]
        Fixed costs of operating each facility
    capacities: numpy array [int]
        Capacities of each facility
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Cornuejols G, Sridharan R, Thizy J-M (1991)
        A Comparison of Heuristics and Relaxations for the Capacitated Plant Location Problem.
        European Journal of Operations Research 50:280-297.
    """
    model = scip.Model(name)

    total_demand = demands.sum()

    model.setMinimize()

    customer_facility_vars = {}
    facility_vars = []
    # add customer-facility vars
    for i, j in itertools.product(range(n_customers), range(n_facilities)):
        var = model.addVar(
            lb=0, ub=1, obj=transportation_cost[i, j], name=f"x_{i}_{j}", vtype="B"
        )
        customer_facility_vars[i, j] = var
    # add facility vars
    for j in range(n_facilities):
        var = model.addVar(lb=0, ub=1, obj=fixed_costs[j], name=f"y_{j}", vtype="B")
        facility_vars.append(var)

    # add constraints
    for i in range(n_customers):
        model.addCons(
            scip.quicksum(customer_facility_vars[i, j] for j in range(n_facilities))
            >= 1
        )
    for j in range(n_facilities):
        model.addCons(
            scip.quicksum(
                demands[i] * customer_facility_vars[i, j] for i in range(n_customers)
            )
            <= capacities[j] * facility_vars[j]
        )

    # optional constraints

    # total capacity constraint
    model.addCons(
        scip.quicksum(capacities[j] * facility_vars[j] for j in range(n_facilities))
        >= total_demand
    )

    # affectation constraints
    for i, j in itertools.product(range(n_customers), range(n_facilities)):
        model.addCons(customer_facility_vars[i, j] <= facility_vars[j])

    return model


def capacitated_warehouse_location(
    n_customers,
    n_facilities,
    transportation_cost,
    demands,
    fixed_costs,
    capacities,
    name="Capacitated Warehouse Location",
):
    """
    Generate a Capacitated Warehouse Location MIP formulation following [1].

    Parameters
    ----------
    n_customers: int
        The desired number of customers
    n_facilities: int
        The desired number of warehouses
    transportation_cost: numpy array [float]
        Matrix of transportation costs from customer i to warehouse j [i,j]
    demands: numpy array [int]
        Demands of each customer
    fixed_costs: numpy array [int]
        Fixed costs of operating each warehouse
    capacities: numpy array [int]
        Capacities of each warehouse
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] J.E. Beasley, An algorithm for solving large capacitated warehouse location problems,
        European Journal of Operational Research, Volume 33, Issue 3, 1988,
        Pages 314-325, ISSN 0377-2217,
        https://doi.org/10.1016/0377-2217(88)90175-0.
    """
    model = scip.Model(name)

    model.setMinimize()

    customer_facility_vars = {}
    facility_vars = []

    # add customer-facility vars
    for i, j in itertools.product(range(n_customers), range(n_facilities)):
        var = model.addVar(
            lb=0, ub=1, obj=transportation_cost[i, j], name=f"x_{i}_{j}", vtype="C"
        )
        customer_facility_vars[i, j] = var

    # add facility vars
    for j in range(n_facilities):
        var = model.addVar(lb=0, ub=1, obj=fixed_costs[j], name=f"y_{j}", vtype="B")
        facility_vars.append(var)

    # add constraints

    # constraints (2)
    for i in range(n_customers):
        model.addCons(
            scip.quicksum(customer_facility_vars[i, j] for j in range(n_facilities))
            >= 1
        )

    # constraints (3)
    for j in range(n_facilities):
        model.addCons(
            scip.quicksum(
                demands[i] * customer_facility_vars[i, j] for i in range(n_customers)
            )
            <= capacities[j] * facility_vars[j]
        )

    # constraints (4) and (5) are skipped because no data of bounds are given in problem data in OR-Library

    # constraints (6)
    for i, j in itertools.product(range(n_customers), range(n_facilities)):
        model.addCons(
            customer_facility_vars[i, j]
            <= min(1, capacities[j] / demands[i]) * facility_vars[j]
        )

    return model



import pyscipopt as scip
from networkx.utils import py_random_state
from geco.mips.set_packing.generic import *


@py_random_state(-1)
def yang_instance(m, seed=0):
    """
    Generates a set packing instance following [1].

    Parameters
    ----------
    m: int
        Number of constraints
    seed: int, random state or None
        Seed for randomizatio.

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
    "Learning Generalized Strong Branching for Set Covering,
    Set Packing, and 0-1 Knapsack Problems", 2020.
    """
    return set_packing(m, *yang_parameters(m, seed), name="Yang Set Packing")


@py_random_state(-1)
def yang_parameters(m, seed=0):
    """
    Generates a set packing instance following [1].

    Parameters
    ----------
    m: int
        Number of constraints
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    n: int
        Number of elements
    values: list[int]
        Value you get for packing each item
    nonzero_vars_for_constraint: list[list[int]]
        Nonzero variables list for each constraint

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
    "Learning Generalized Strong Branching for Set Covering,
    Set Packing, and 0-1 Knapsack Problems", 2020.
    """
    n = 5 * m
    values = [seed.randint(1, 100) for _ in range(n)]
    num_nonzero_vars_for_constraint = [
        seed.randint(2 * n // 25 + 1, 3 * n // 25 - 1) for _ in range(m)
    ]
    nonzero_vars_for_constraint = [
        seed.sample(range(n), k=num) for num in num_nonzero_vars_for_constraint
    ]
    return n, values, nonzero_vars_for_constraint



import tempfile
from urllib.request import urlretrieve, urlopen
from urllib.error import URLError
import pyscipopt as scip
import os
import pandas as pd


class Loader:
    def __init__(self, persistent_directory=None):
        """
        Initializes the MIPLIB loader object

        Parameters
        ----------
        persistent_directory: str or None
            Path for directory to use for persistent files,
            If set to None, resorts to default case of using temporary files
            that get deleted after program execution
        """
        self.instances_cache = {}
        self.dir = persistent_directory
        if persistent_directory:
            self._load_instances_cache()

    def load_instance(self, instance_name, with_solution=False):
        if not self._instance_cached(instance_name):
            self._download_instance(instance_name)
        problem_path = self._instance_path(instance_name)
        model = scip.Model()
        model.readProblem(problem_path)
        if with_solution:
            self._add_solution(model, instance_name)
        return model

    def _instance_path(self, instance_name):
        return self.instances_cache[instance_name]

    def _generate_path_for_instance(self, instance_name):
        if self.dir:
            return self.dir + instance_name
        else:
            extension = instance_name[instance_name.index(".") :]
            return tempfile.NamedTemporaryFile(suffix=extension, delete=False).name

    def _download_instance(self, instance_name):
        path = self._generate_path_for_instance(instance_name)
        url = self._look_for_working_url(self._instance_urls(instance_name))
        if url:
            urlretrieve(url, path)
            self.instances_cache[instance_name] = path
        else:
            raise ValueError(
                "Was not able to find the instance in any of the MIPLIB sources"
            )

    def _look_for_working_url(self, urls):
        for url in urls:
            try:
                response = urlopen(url)
            except URLError:
                continue
            if self._successful_response(response):
                return url
        return None

    @staticmethod
    def _successful_response(response):
        return response.status == 200 and "not_found" not in response.url

    def _instance_cached(self, instance_name):
        return instance_name in self.instances_cache

    def _load_instances_cache(self):
        for path in os.listdir(self.dir):
            if path.endswith(".mps.gz"):
                instance_name = path.split("/")[-1]
                self.instances_cache[instance_name] = self.dir + path

    def _add_solution(self, model, instance_name):
        url = self._look_for_working_url(self._solution_urls(instance_name))
        if url:
            with tempfile.NamedTemporaryFile(suffix=".sol.gz") as sol_file:
                urlretrieve(url, sol_file.name)
                model.readSol(sol_file.name)
        else:
            raise ValueError(
                "Was not able to find the solution in any of the MIPLIB sources"
            )

    @staticmethod
    def _instance_urls(instance_name):
        return [
            f"https://miplib.zib.de/WebData/instances/{instance_name}",  # 2017 instances
            f"http://miplib2010.zib.de/download/{instance_name}",  # 2010 instances
            f"http://miplib2010.zib.de/miplib2003/download/{instance_name}",  # 2003 instance
        ]

    @staticmethod
    def _solution_urls(instance_name):
        name = instance_name[: instance_name.index(".")]
        return [
            f"https://miplib.zib.de/downloads/solutions/{name}/1/{name}.sol.gz",  # 2017 solutions
            f"http://miplib2010.zib.de/download/{name}.sol.gz",  # 2010 solutions
            f"http://miplib2010.zib.de/miplib2003/download/{name}.sol.gz",  # 2003 solutions
        ]

    def __del__(self):
        if self.dir is None:
            for path in self.instances_cache.values():
                os.unlink(path)


def benchmark_instances():
    for instance in custom_list("https://miplib.zib.de/downloads/benchmark-v2.test"):
        yield instance


def easy_instances():
    for instance in custom_list("https://miplib.zib.de/downloads/easy-v9.test"):
        yield instance


def hard_instances():
    for instance in custom_list("https://miplib.zib.de/downloads/hard-v15.test"):
        yield instance


def open_instances():
    for instance in custom_list("https://miplib.zib.de/downloads/open-v14.test"):
        yield instance


def custom_list(source, with_solution=False, loader=None):
    """
    Returns a generator of instances from the given list

    Parameters
    ----------
    source: str
        Path or URL for the instance list source
    with_solution: bool
        Whether to return the instance with the known solutions or not
    loader: Loader
        Loader object to download instances with

    Returns
    -------
    A generator for the instances
    """
    df = pd.read_csv(source, names=["instance"])
    if loader is None:
        loader = Loader()
    for instance in df["instance"]:
        yield loader.load_instance(instance, with_solution=with_solution)



import math

from networkx.utils import py_random_state

from geco.mips.knapsack.generic import knapsack


@py_random_state(-1)
def yang_params(n, seed=0):
    """
    Generates knapsack instance params according to [1].

    Parameters
    ----------
    n: int
        Number of items
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    profits: list[int]
        Profit of each item
    weights: list[int]
        Weight of each item
    capacity: float
        Capacity of knapsack

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
        "Learning Generalized Strong Branching for Set Covering,
        Set Packing, and 0-1 Knapsack Problems", 2020.
    """

    def draw_value():
        return seed.randint(1, 10 * n)

    profits = [draw_value() for _ in range(n)]
    weights = [draw_value() for _ in range(n)]
    capacity = math.floor(sum(weights) / 5)

    return profits, weights, capacity


@py_random_state(-1)
def yang_instance(n, seed=0):
    """
    Generates knapsack instance according to [1].

    Parameters
    ----------
    n: int
        Number of items
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
        "Learning Generalized Strong Branching for Set Covering,
        Set Packing, and 0-1 Knapsack Problems", 2020.
    """
    return knapsack(*yang_params(n, seed))



import networkx as nx
import pyscipopt as scip

from geco.mips.independent_set.generic import clique_independent_set


def gasse_params(n, p, seed=0):
    """
    Generates a maximum independent set instance as described in [1].

    Parameters
    ----------
    n: int
        Number of nodes.
    p: float
        Edge probability
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    graph: nx.Graph
        An Erdos-Renyi Networkx graph

    References
    ----------
    .. [1] "Exact Combinatorial Optimization with Graph Convolutional Neural Networks" (2019)
      Maxime Gasse, Didier Chételat, Nicola Ferroni, Laurent Charlin and Andrea Lodi
      Advances in Neural Information Processing Systems 32 (2019)
    """
    return nx.generators.erdos_renyi_graph(n, p, seed)


def gasse_instance(n, p, seed=0):
    """
    Generates a maximum independent set instance as described in [1].

    Parameters
    ----------
    n: int
        Number of nodes.
    p: float
        Edge probability
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the instance.

    References
    ----------
    .. [1] "Exact Combinatorial Optimization with Graph Convolutional Neural Networks" (2019)
      Maxime Gasse, Didier Chételat, Nicola Ferroni, Laurent Charlin and Andrea Lodi
      Advances in Neural Information Processing Systems 32 (2019)
    """
    return clique_independent_set(
        gasse_params(n, p, seed), name="Gasse Independent Set"
    )



import itertools
import types

import pytest

from geco.mips.knapsack.pisinger import *
from geco.mips.knapsack.yang import *

"""
Yang Generators Tests
"""


@pytest.mark.parametrize(
    "n,seed", itertools.product([10, 100, 1000], [0, 1, 1337, 53115])
)
def test_yang_knapsack_creation(n, seed):
    model = yang_instance(n, seed=seed)
    assert model.getNVars() == n
    assert model.getNConss() == 1
    assert model.getObjectiveSense() == "maximize"


def test_yang_knapsack_solution_1():
    model = knapsack([1], [1], 0)
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 0


def test_yang_knapsack_solution_2():
    model = knapsack([1, 1, 1], [1, 1, 1], 5)
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 3


@pytest.mark.parametrize(
    "n,seed1,seed2",
    itertools.product([3, 10, 15], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_yang_seeding(n, seed1, seed2):
    params1 = yang_params(n, seed=seed1)
    params2 = yang_params(n, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


"""
Pisinger Generators Tests
"""


def test_pisinger_creation_of_all():
    n = 100
    c = 20
    models = [
        uncorrelated(n, c),
        weakly_correlated(n, c),
        strongly_correlated(n, c),
        inverse_strongly_correlated(n, c),
        almost_strongly_correlated(n, c),
        subset_sum(n, c),
        uncorrelated_with_similar_weights(n, c),
        profit_ceiling(n, c),
        circle(n, c),
        multiple_strongly_correlated(n, c, 7, 10, 3),
        spanner(10, 100, 100, uncorrelated_distribution, 500),
    ]
    for model in models:
        assert model.getNVars() == n
        assert model.getNConss() == 1
        assert model.getObjectiveSense() == "maximize"


@pytest.mark.parametrize(
    "n,R,distribution,seed1,seed2",
    itertools.product(
        [3, 10, 15],
        [10, 100],
        [
            uncorrelated_distribution,
            weakly_correlated_distribution,
            strongly_correlated_distribution,
            inverse_strongly_correlated_distribution,
            almost_strongly_correlated_distribution,
            subset_sum_distribution,
            uncorrelated_with_similar_weights_distribution,
        ],
        [0, 1, 1337, 53115],
        [0, 1, 1337, 53115],
    ),
)
def test_pisinger_seeding(n, R, distribution: types.FunctionType, seed1, seed2):
    params1 = generate_from_distribution(n, **distribution(R=R, seed=seed1))
    params2 = generate_from_distribution(n, **distribution(R=R, seed=seed2))
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



import itertools

import pytest

from geco.mips.production_planning.tang import *


def test_tang():
    T = 100
    model = tang_instance(T)
    assert model.getNVars() == (T + 1) * 3 - 1
    assert model.getNConss() == 2 * T + 2
    assert model.getObjectiveSense() == "minimize"


def test_tang_simple_feasible():
    T = 1
    initial_storage, final_storage, p, h, q, d = 0, 20, [1, 1], [1, 1], [1, 1], [0, 0]
    M = 20
    params = M, initial_storage, final_storage, p, h, q, d
    model = uncapacitated_lot_sizing(T, *params)
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 20 + 20 + 1


def test_tang_simple_infeasible():
    T = 1
    initial_storage, final_storage, p, h, q, d = 0, 20, [1, 1], [1, 1], [1, 1], [0, 0]
    M = 19
    params = M, initial_storage, final_storage, p, h, q, d
    model = uncapacitated_lot_sizing(T, *params)
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "infeasible"


@pytest.mark.parametrize(
    "T,seed1,seed2",
    itertools.product([10, 100, 200], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_seeding(T, seed1, seed2):
    params1 = tang_params(T, seed=seed1)
    params2 = tang_params(T, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



import pandas as pd
import pytest
import itertools

from geco.mips.loading.miplib import *


def test_load_miplib_list():
    df = pd.read_csv("data/lists/branching_rules_revisited.csv", comment="#")
    loader = Loader()
    miplib_only_df = df[df["miplib"] == 1]
    for i in miplib_only_df["instance"]:
        loader.load_instance(f"{i}.mps.gz")
        path = loader.instances_cache[f"{i}.mps.gz"]
        assert os.path.exists(path)


def test_load_instance():
    instance = Loader().load_instance("30n20b8.mps.gz")
    assert isinstance(instance, scip.Model)


def test_deletion_of_temp_files():
    loader = Loader()
    instance_name = "30n20b8.mps.gz"
    loader.load_instance(instance_name)
    path = loader.instances_cache[instance_name]
    del loader
    assert not os.path.exists(path)


def test_persistent_directory():
    loader = Loader(persistent_directory="./")
    instance_name = "30n20b8.mps.gz"
    loader.load_instance(instance_name)
    del loader
    new_loader = Loader(persistent_directory="./")
    path = new_loader.instances_cache[instance_name]
    assert (
        instance_name in new_loader.instances_cache
    )  # instance path loaded correctly into cache
    assert os.path.exists(path)  # instance path exists
    os.unlink(instance_name)  # cleanup local directory


def test_miplib_sources_with_solution():
    _check_instance("30n20b8.mps.gz", with_solution=True)  # from loading 2017
    _check_instance("neos-941262.mps.gz", with_solution=True)  # from loading 2010
    _check_instance("vpm2.mps.gz", with_solution=True)  # from loading 2003


def test_instance_not_found_error():
    with pytest.raises(ValueError):
        Loader().load_instance("i_am_not_an_instance.mps.gz")


def test_solution_not_found_error():
    with pytest.raises(ValueError):
        Loader().load_instance(
            "bharat.mps.gz", with_solution=True
        )  # one of loading 2017 open instances with no known solution


def _check_instance(instance_name, with_solution=False):
    instance = Loader().load_instance(instance_name, with_solution=with_solution)
    assert isinstance(instance, scip.Model)
    if with_solution:
        sols = instance.getSols()
        assert len(sols) == 1


def test_custom_list():
    miplib_2017_list = "https://miplib.zib.de/downloads/collection-v1.test"
    for instance in itertools.islice(custom_list(miplib_2017_list), 5):
        assert isinstance(instance, scip.Model)


def test_easy_instances():
    for instance in itertools.islice(easy_instances(), 5):
        assert isinstance(instance, scip.Model)


def test_hard_instances():
    for instance in itertools.islice(hard_instances(), 5):
        assert isinstance(instance, scip.Model)


def test_open_instances():
    for instance in itertools.islice(open_instances(), 5):
        assert isinstance(instance, scip.Model)


def test_benchmark_instances():
    for instance in itertools.islice(benchmark_instances(), 5):
        assert isinstance(instance, scip.Model)



import pyscipopt as scip


def packing(n, m, costs, constraint_coefficients, limits, binary, name="Packing"):
    """Generates a packing instance as described in A.2 in [1].

    Parameters:
    ----------
    n: int
        Number of variables
    m: int
        Number of constraints
    costs: list[number] of size n
        Coefficients of objective function
    constraint_coefficients: list[list[number]] of dimensions (m x n)
        Coefficients of each variable for each constraint
    limits: list[number] of size m
        Limits of each constraint
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References:
    ----------
    .. [1] Tang, Y., Agrawal, S., & Faenza, Y. (2019). Reinforcement learning for integer
    programming: Learning to cut. arXiv preprint arXiv:1906.04859.
    """
    model = scip.Model(name)

    # add variables and their cost
    vars = []
    for i in range(n):
        cost = costs[i]
        if binary:
            var = model.addVar(lb=0, ub=1, obj=cost, name=f"v_{i}", vtype="B")
        else:
            var = model.addVar(lb=0, ub=None, obj=cost, name=f"v_{i}", vtype="I")
        vars.append(var)

    # add constraints
    for i in range(m):
        constraint_vars = (
            constraint_coefficients[i][j] * var for j, var in enumerate(vars)
        )
        model.addCons(scip.quicksum(constraint_vars) <= limits[i])

    # contrary to the paper (as of 05/11/2020) as a result of correspondence of with one of the authors
    model.setMaximize()

    return model



import tempfile

from geco.mips.set_cover.yang import yang_instance
from geco.mips.utilities.generic import *


def test_saving_shuffled_instance():
    model = yang_instance(300)
    shuffled_model = shuffle(model, seed=1)

    original_file = tempfile.NamedTemporaryFile(suffix=".mps")
    shuffled_file = tempfile.NamedTemporaryFile(suffix=".mps")

    model.writeProblem(original_file.name)
    shuffled_model.writeProblem(shuffled_file.name)

    with open(original_file.name, "r") as of, open(shuffled_file.name, "r") as sf:
        for line1, line2 in zip(of, sf):
            if line1 != line2:
                assert True
                break
        else:
            assert False


def test_expand_parameters():
    def add(x, y):
        return x + y

    assert list(expand_parameters(add, x=[1, 2], y=[3, 4])) == [4, 5, 5, 6]



import networkx as nx
import math


def edgeweight_properties(graph):
    """
    Calculates properties of edge weights.

    Parameters
    ----------
    graph: nx.Graph
        Graph to calculate the properties from

    Returns
    -------
    max_weight: number
        Maximum weights of an edge
    min_weight: number
        Minimum weight of an edge
    num_of_zero_weights: int
        Number of edges with zero weight
    """
    max_weight = -math.inf
    min_weight = math.inf
    num_of_zero_weights = 0
    for u, v, d in graph.edges(data=True):
        weight = d["weight"]
        if weight > max_weight:
            max_weight = weight
        if weight < min_weight:
            min_weight = weight
        if weight == 0:
            num_of_zero_weights += 1

    return max_weight, min_weight, num_of_zero_weights


def graph_properties(g):
    """
    Calculates properties of edge weights.

    Parameters
    ----------
    graph: nx.Graph
        Graph to calculate the properties from

    Returns
    -------
    num_nodes: int
        Number of nodes of the graph
    num_edges: int
        Number of edges of the graph
    avg_degree: float
        Average degree of node
    density: float
        Number of edges divided by the number of possible edges
    planar: bool
        Whether the graph is planar or not
    max_edgeweight: float
        Maximum weights of an edge
    min_edgeweight: float
        Minimum weight of an edge
    num_of_zero_weights: int
        Number of edges with zero weight
    num_of_connected_components: int
        Number of connected components in the graph
    max_degree: int
        Maximum degree of a node in the graph
    assortativity_coeff: float
        The degree assortativity coefficient as defined in [1]
    number_of_triangles: int
        Number of traingles in the graph
    average_clustering_coeff: float
        Average clustering coefficient as defined in [2]
    max_k_core: int
        Maximum k-core as defined in [3]

    References
    ----------
    ..[1] https://networkx.org/documentation/stable//reference/algorithms/generated/networkx.algorithms.assortativity.degree_assortativity_coefficient.html
    ..[2] https://networkx.org/documentation/networkx-2.4/reference/algorithms/generated/networkx.algorithms.cluster.average_clustering.html
    ..[3] https://networkx.org/documentation/networkx-1.9/reference/generated/networkx.algorithms.core.core_number.html
    """

    avg_degree = sum((deg for node, deg in g.degree)) / g.number_of_nodes()
    density = g.number_of_edges() / (
        (g.number_of_nodes() * g.number_of_nodes() - g.number_of_nodes()) / 2
    )
    planar, _ = nx.algorithms.check_planarity(g)
    max_edgeweight, min_edgeweight, num_of_zero_edgeweights = edgeweight_properties(g)
    num_of_connected_components = nx.number_connected_components(g)
    number_of_triangles = sum(nx.triangles(g).values()) // 3
    max_degree = max(degree for _, degree in g.degree())
    assortativity_coeff = nx.degree_assortativity_coefficient(g)
    average_clustering_coeff = nx.average_clustering(g)
    max_k_core = max(nx.core_number(g).values())

    return {
        "num_nodes": g.number_of_nodes(),
        "num_edges": g.number_of_edges(),
        "avg_degree": avg_degree,
        "density": density,
        "planar": planar,
        "max_edgeweight": max_edgeweight,
        "min_edgeweight": min_edgeweight,
        "num_of_zero_edgeweights": num_of_zero_edgeweights,
        "num_of_connected_components": num_of_connected_components,
        "max_degree": max_degree,
        "assortativity_coeff": assortativity_coeff,
        "number_of_triangles": number_of_triangles,
        "average_clustering_coeff": average_clustering_coeff,
        "max_k_core": max_k_core,
    }



from networkx.utils import py_random_state

from geco.mips.set_cover.generic import set_cover


@py_random_state(-1)
def yang_instance(m, seed=0):
    """
    Generates instance for set cover generation as described in [1].

    Parameters
    ----------
    m: int
        Number of set constraints
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
    "Learning Generalized Strong Branching for Set Covering,
    Set Packing, and 0-1 Knapsack Problems", 2020.
    """
    return set_cover(*yang_params(m, seed))


@py_random_state(-1)
def yang_params(m, seed=0):
    """
    Generates instance params for set cover generation as described in [1].

    Parameters
    ----------
    m: int
        Number of set constraints
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    costs: list[int]
        Element costs in objective function
    sets: list[set]
        Definition of element requirement for each set

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
    "Learning Generalized Strong Branching for Set Covering,
    Set Packing, and 0-1 Knapsack Problems", 2020.
    """
    n = 10 * m

    costs = [seed.randint(1, 100) for _ in range(n)]

    sets = []
    for _ in range(m):
        num_nonzero = seed.randint(2 * n // 25 + 1, 3 * n // 25 - 1)
        sets.append(set(j for j in seed.sample(range(n), k=num_nonzero)))

    return costs, sets



import itertools
import pyscipopt as scip

import geco.mips.utilities.naming as naming


def naive(graph):
    model = scip.Model("Naive MaxCut")

    node_variables = {}
    for v in graph.nodes():
        node_variables[v] = model.addVar(lb=0, ub=1, obj=0, name=str(v), vtype="B")

    edge_variables = {}
    all_non_negative = True
    for u, v, d in graph.edges(data=True):
        edge_name = naming.undirected_edge_name(u, v)
        weight = d["weight"]
        edge_variables[edge_name] = model.addVar(
            lb=0, ub=1, obj=weight, name=edge_name, vtype="B"
        )
        if weight < 0:
            all_non_negative = False

    model.setMaximize()

    for u, v, d in graph.edges(data=True):
        edge_name = naming.undirected_edge_name(u, v)
        model.addCons(
            node_variables[u] + node_variables[v] + edge_variables[edge_name] <= 2
        )
        model.addCons(
            -node_variables[u] - node_variables[v] + edge_variables[edge_name] <= 0
        )
        if not all_non_negative:
            model.addCons(
                node_variables[u] - node_variables[v] - edge_variables[edge_name] <= 0
            )
            model.addCons(
                -node_variables[u] + node_variables[v] - edge_variables[edge_name] <= 0
            )

    return (node_variables, edge_variables), model


def triangle(graph):
    model = scip.Model("Triangle MaxCut")

    edge_variables = {}

    for u, v in itertools.combinations(graph.nodes(), 2):
        edge_name = naming.undirected_edge_name(u, v)
        if graph.has_edge(u, v):
            weight = graph.get_edge_data(u, v)["weight"]
        else:
            weight = 0
        edge_variables[edge_name] = model.addVar(
            lb=0, ub=1, obj=weight, name=edge_name, vtype="B"
        )

    model.setMaximize()

    for i, j, k in itertools.combinations(graph.nodes(), 3):
        x_ij = _get_edge_variable(i, j, edge_variables)
        x_ik = _get_edge_variable(i, k, edge_variables)
        x_kj = _get_edge_variable(k, j, edge_variables)
        model.addCons(x_ij <= x_ik + x_kj)
        model.addCons(x_ij + x_ik + x_kj <= 2)

    return edge_variables, model


def _get_edge_variable(u, v, edge_variables):
    edge_name = naming.undirected_edge_name(u, v)
    return edge_variables[edge_name]



import pyscipopt as scip

from geco.mips.loading.orlib import *
from geco.mips.set_cover.generic import set_cover
from geco.mips.loading.orlib import orlib_load_instance
from geco.mips.set_cover.generic import set_cover


def scp_reader(file):
    """
    Reads scp set-cover instances mentioned in [1].

    Parameters
    ----------
    file: file-like object

    Returns
    -------
    costs: list[int]
        Element costs in objective function
    sets: list[set]
        Definition of element requirement for each set

    References
    ----------
    ..[1] http://people.brunel.ac.uk/~mastjjb/jeb/orlib/scpinfo.html
    """

    number_of_cons, number_of_vars = read_numbers(file.readline())
    costs = read_multiline_numbers(file, number_of_vars)
    sets = []
    while file:
        number_of_vars_in_constraint = read_number(file.readline())
        if not number_of_vars_in_constraint:
            break
        constraint = list(read_multiline_numbers(file, number_of_vars_in_constraint))
        constraint = zero_index(constraint)
        sets.append(constraint)
    assert len(costs) == number_of_vars and len(sets) == number_of_cons
    return costs, sets


def rail_reader(file):
    """
    Reads rail set-cover instances mentioned in [1].

    Parameters
    ----------
    file: file-like object

    Returns
    -------
    costs: list[int]
        Element costs in objective function
    sets: list[set]
        Definition of element requirement for each set

    References
    ----------
    ..[1] http://people.brunel.ac.uk/~mastjjb/jeb/orlib/scpinfo.html
    """

    number_of_cons, number_of_vars = read_numbers(file.readline())

    costs = []
    sets = [[] for _ in range(number_of_cons)]
    col_idx = 0
    while file:
        line = file.readline()
        if not line:
            break

        numbers = list(read_numbers(line))
        costs.append(numbers[0])
        rows_covered = zero_index(numbers[2:])

        for row in rows_covered:
            sets[row].append(col_idx)
        col_idx += 1
    sets = list(filter(lambda l: len(l) > 0, sets))
    assert len(costs) == number_of_vars and len(sets) == number_of_cons
    return costs, sets


def orlib_instance(instance_name):
    """
    Loads an orlib Set-cover instance

    Parameters
    ----------
    instance_name: str
        Name of the set-cover file. example: "scp41.txt"

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the loaded instance
    """
    # TODO: assert that instance_name correlated to one of the listed set-cover files
    if instance_name[:3] == "scp":
        return orlib_load_instance(
            instance_name, reader=scp_reader, formulation=set_cover
        )
    elif instance_name[:4] == "rail":
        return orlib_load_instance(
            instance_name, reader=rail_reader, formulation=set_cover
        )



import itertools

import pytest

from geco.mips.set_packing.yang import *


def test_yang():
    m = 100
    model = yang_instance(m)
    assert model.getNVars() == 5 * m
    assert model.getNConss() == m
    assert model.getObjectiveSense() == "maximize"


def test_yang_simple_instance():
    params = m, n, costs, nonzero_vars_for_constraints = 1, 1, [5], [[0]]
    model = set_packing(*params)
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 5


@pytest.mark.parametrize(
    "m,seed1,seed2",
    itertools.product([10, 100], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_yang_seeding(m, seed1, seed2):
    params1 = yang_parameters(m, seed=seed1)
    params2 = yang_parameters(m, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



import pytest

from geco.mips.facility_location.cornuejols import *
from geco.mips.facility_location.generic import *
from geco.mips.facility_location.orlib import *


def test_capacitated_facility_location():
    n_customers, n_facilities, ratio = 25, 10, 2
    instance_params = cornuejols_params(n_customers, n_facilities, ratio, seed=0)
    model = capacitated_facility_location(n_customers, n_facilities, *instance_params)
    assert model.getNVars() == n_customers * n_facilities + n_facilities
    assert (
        model.getNConss() == n_customers + n_facilities + 1 + n_customers * n_facilities
    )
    assert model.getObjectiveSense() == "minimize"
    model.hideOutput()
    model.optimize()
    assert 5403 <= model.getObjVal() <= 5404


@pytest.mark.parametrize(
    "n_customers,n_facilities,ratio,seed1,seed2",
    itertools.product(
        [3, 10, 15], [3, 10, 15], [1, 2], [0, 1, 1337, 53115], [0, 1, 1337, 53115]
    ),
)
def test_seeding(n_customers, n_facilities, ratio, seed1, seed2):
    params1 = cornuejols_params(n_customers, n_facilities, ratio, seed=seed1)
    params2 = cornuejols_params(n_customers, n_facilities, ratio, seed=seed2)
    something_different = False
    for param1, param2 in zip(params1, params2):
        if (param1 != param2).any():
            something_different = True
            break
    same_seeds_produce_same_params = seed1 == seed2 and not something_different
    different_seeds_produce_different_params = seed1 != seed2 and something_different
    assert same_seeds_produce_same_params or different_seeds_produce_different_params


def test_orlib_cap_numeric():
    instance_name = "cap41.txt"
    n_customers, n_warehouses = 50, 16
    instance = orlib_instance(instance_name)
    assert instance.getNVars() == n_warehouses * n_customers + n_warehouses


def test_orlib_cap_alpha():
    instance_name = "capa2.txt"
    n_customers, n_warehouses = 1000, 100
    instance = orlib_instance(instance_name)
    assert instance.getNVars() == n_warehouses * n_customers + n_warehouses


def test_orlib_wrong_instance_name():
    with pytest.raises(ValueError):
        orlib_instance("asdlkfj.txt")


@pytest.mark.parametrize(
    "instance_name",
    [
        "cap41",
        "cap51",
        "cap61",
        "cap71",
        "cap81",
        "cap91",
        "cap101",
        "cap111",
        "cap121",
        "cap131",
    ],
)
def test_orlib_solution(instance_name):
    import pandas as pd

    df = pd.read_csv(
        "data/lists/orlib_capacitated_warehouse_location_solutions.csv", comment="#"
    )

    instance = orlib_instance(instance_name + ".txt")
    instance.hideOutput()
    instance.optimize()
    assert instance.getStatus() == "optimal"
    assert pytest.approx(
        instance.getObjVal() == df[df["name"] == instance_name]["solution_value"]
    )



import pyscipopt as scip


def set_packing(m, n, values, nonzero_vars_for_constraint, name="Set Packing"):
    """
    Generates a set packing formulation following [1].

    Parameters
    ----------
    m: int
        Number of constraints
    n: int
        Number of elements
    values: list[int]
        Value you get for packing each item
    nonzero_vars_for_constraint: list[list[int]]
        Nonzero variables list for each constraint
    name: str
        Name of the model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Yu Yang, Natashia Boland, Bistra Dilkina, Martin Savelsbergh,
    "Learning Generalized Strong Branching for Set Covering,
    Set Packing, and 0-1 Knapsack Problems", 2020.
    """
    model = scip.Model(name)

    # add variables and their cost
    vars = []
    for i in range(n):
        var = model.addVar(lb=0, ub=1, obj=values[i], name=f"v_{i}", vtype="B")
        vars.append(var)

    # add constraints
    for i in range(m):
        nonzero_vars = (vars[j] for j in nonzero_vars_for_constraint[i])
        model.addCons(scip.quicksum(nonzero_vars) <= 1)

    model.setMaximize()

    return model



import itertools

import pytest

from geco.mips.packing.tang import *


def test_tang_integral():
    n, m = 10, 10
    model = tang_instance(n, m, binary=False)
    vars = model.getVars()
    assert len(vars) == n
    assert all(var.vtype() == "INTEGER" for var in vars)
    assert model.getNConss() == m
    assert model.getObjectiveSense() == "maximize"


def test_tang_binary():
    n, m = 10, 10
    model = tang_instance(n, m, binary=True)
    vars = model.getVars()
    assert len(vars) == n
    assert all(var.vtype() == "BINARY" for var in vars)
    assert model.getNConss() == m
    assert model.getObjectiveSense() == "maximize"


def test_tang_simple_instance():
    n, m, costs, constraint_coefficients, limits, binary = 1, 1, [1], [[1]], [5], False
    model = packing(n, m, costs, constraint_coefficients, limits, binary)
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 5


@pytest.mark.parametrize(
    "n,m,binary,seed1,seed2",
    itertools.product(
        [3, 10, 100],
        [3, 10, 100],
        [True, False],
        [0, 1, 1337, 53115],
        [0, 1, 1337, 53115],
    ),
)
def test_seeding(n, m, binary, seed1, seed2):
    params1 = tang_params(n, m, binary, seed=seed1)
    params2 = tang_params(n, m, binary, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and params1 == params2
    different_seeds_produce_different_params = seed1 != seed2 and params1 != params2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



from networkx.utils import py_random_state

from geco.mips.set_cover.generic import set_cover


def _sun_costs(n, seed):
    return [seed.randint(1, 100) for _ in range(n)]


def _sun_sets(n, m, seed, initial_sets=None):
    if not initial_sets:
        sets = [set() for _ in range(m)]
    else:
        sets = list(initial_sets)

    p = 0.05
    for e in range(n):
        # enforce element to appear in at least 2 sets
        for s in (sets[i] for i in seed.sample(range(m), k=2)):
            s.add(e)

        # add element to set with probability p
        for s in sets:
            if seed.random() < p:
                s.add(e)

    return sets


@py_random_state(-1)
def sun_instance(n, m, seed=0):
    """
    Generates instance for set cover generation as described in [1].

    Parameters
    ----------
    n: int
        Number of elements
    m: int
        Number of set constraints
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    References
    ----------
    .. [1] Haoran Sun, Wenbo Chen, Hui Li, & Le Song (2021).
         Improving Learning to Branch via Reinforcement Learning. In Submitted to
         International Conference on Learning
    """
    return set_cover(*sun_params(n, m, seed))


@py_random_state(-1)
def sun_params(n, m, seed=0):
    """
    Generates instance params for set cover generation as described in [1].

    Parameters
    ----------
    n: int
        Number of elements
    m: int
        Number of set constraints
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    costs: list[int]
        Element costs in objective function
    sets: list[set]
        Definition of element requirement for each set

    References
    ----------
    .. [1] Haoran Sun, Wenbo Chen, Hui Li, & Le Song (2021).
         Improving Learning to Branch via Reinforcement Learning. In Submitted to
         International Conference on Learning
    """
    return _sun_costs(n, seed), _sun_sets(n, m, seed, initial_sets=None)


@py_random_state(-1)
def expand_sun_params(new_params, base_result, seed=0):
    """
    Implements the expansion from an existing set cover instance as described in [1].

    Parameters
    ----------
    new_params: tuple
        New params for sun_params
    base_result: tuple
        Tuple of (costs, sets) that represent instance params of backbone
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    costs: list[int]
        Element costs in objective function
    sets: list[set]
        Definition of element requirement for each set

    References
    __________
    .. [1] Haoran Sun, Wenbo Chen, Hui Li, & Le Song (2021).
     Improving Learning to Branch via Reinforcement Learning. In Submitted to
     International Conference on Learning
    """
    n, *_ = new_params
    base_costs, base_sets = base_result
    assert n > len(base_costs)

    costs = list(base_costs)
    costs += _sun_costs(n - len(base_costs), seed)

    return costs, _sun_sets(n, len(base_sets), seed, initial_sets=base_sets)



import pytest
import os
from geco.mips.loading.orlib import *


def test_read_multiline_correct():
    content = b"""1 2 3.5\n 3 2 1\n"""
    with open("temp", "wb") as f:
        f.write(content)

    with open("temp", "rb") as f:
        numbers = read_multiline_numbers(file=f, number_to_read=6)
        assert len(numbers) == 6
        assert list(numbers) == [1, 2, 3.5, 3, 2, 1]

    with open("temp", "rb") as f, pytest.raises(ValueError):
        read_multiline_numbers(f, 5)

    os.unlink("temp")


def test_read_number():
    n_1 = read_number(b"1")
    n_half = read_number(b".5")
    assert isinstance(n_1, int) and n_1 == 1
    assert isinstance(n_half, float) and n_half == 0.5


def test_read_numbers():
    expected_numbers = [1, 2.5, 3]
    numbers = read_numbers(b"  1 2.5 3 \n")
    for read, expected in zip(numbers, expected_numbers):
        assert read == expected



import pyscipopt as scip


def set_cover(costs, sets, name="Set Cover"):
    """
    Generates basic set cover formulation.

    Parameters
    ----------
    costs: list[float]
        Cost for covering each element
    sets: list[set]
        Set constraints for elements

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the generated instance

    """
    model = scip.Model(name)

    # add variables and their cost
    variables = [
        model.addVar(lb=0, ub=1, obj=c, name=f"v_{i}", vtype="B")
        for i, c in enumerate(costs)
    ]

    # add constraints
    for s in sets:
        model.addCons(scip.quicksum(variables[i] for i in s) >= 1)

    model.setMinimize()

    return model



def undirected_edge_name(u, v) -> str:
    """
    :return The name of an undirected edge as "(u,v)" with u <= v.
    """
    u_i, v_i = int(u), int(v)
    if u_i > v_i:
        u_i, v_i = v_i, u_i
    return f"({u_i},{v_i})"



import dwave_networkx as dwave
from networkx import write_weighted_edgelist
from random import random
import numpy as np
from networkx.utils import py_random_state


def draw_intra_weight(seed):
    return seed.random() - 0.5


def draw_inter_weight(seed):
    return seed.random() * 2 - 1


def _initialize_weights_pegasus(
    pegasus_graph, size, draw_inter_weight, draw_intra_weight, draw_other_weight
):
    # 'nice' coordinate indices names (as per d-wave's documentation):
    p_coor = dwave.pegasus_coordinates(size)

    for _from, _to in pegasus_graph.edges:
        _from_nice = p_coor.linear_to_nice(_from)
        _to_nice = p_coor.linear_to_nice(_to)
        if not in_chimera_subgraph(_from_nice, _to_nice):
            pegasus_graph.add_edge(_from, _to, weight=draw_other_weight())
        else:
            if in_same_chimera_tile(_from_nice, _to_nice):
                # edge from one side to the other (internal edge)
                if not on_same_side(_from_nice, _to_nice):
                    pegasus_graph.add_edge(_from, _to, weight=draw_intra_weight())
                else:  # odd couplers
                    pegasus_graph.add_edge(_from, _to, weight=draw_other_weight())
            else:
                pegasus_graph.add_edge(_from, _to, weight=draw_inter_weight())


def in_chimera_subgraph(_from_nice, _to_nice):
    t = 0
    return _from_nice[t] + _to_nice[t] == 0


def on_same_side(_from_nice, _to_nice):
    u = 3
    return _from_nice[u] == _to_nice[u]


def in_same_chimera_tile(_from_nice, _to_nice):
    y, x = 1, 2
    return _from_nice[y] == _to_nice[y] and _from_nice[x] == _to_nice[x]


@py_random_state(-1)
def dwave_pegasus_graph(
    size,
    draw_inter_weight=draw_inter_weight,
    draw_intra_weight=draw_intra_weight,
    draw_other_weight=draw_inter_weight,
    seed=0,
):
    """
    Generate DWave Pegasus graph as described in [1] using dwave_networkx.

    Parameters
    ----------
    size: int
       Size of generated pegasus graph
    draw_inter_weight: function (seed) -> number
        Function to call for weights of inter-cell edges
    draw_intra_weight: function (seed) -> number
        Function to call for weights of intra-cell edges
    draw_other_weight: function (seed) -> number
        Function to call for weights of intra-cell edges

    Returns
    -------
    graph: nx.Graph
        The generated Pegasus graph

    References
    ----------
    ..[1] https://docs.ocean.dwavesys.com/en/latest/concepts/topology.html
    """
    g = dwave.pegasus_graph(size)
    _initialize_weights_pegasus(
        pegasus_graph=g,
        size=size,
        draw_inter_weight=lambda: draw_inter_weight(seed),
        draw_intra_weight=lambda: draw_intra_weight(seed),
        draw_other_weight=lambda: draw_other_weight(seed),
    )
    return g



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    readme_text = f.read()


setup(
    name="GeCO",
    version="1.0.5",
    description="Generators for Combinatorial Optimization",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    url="https://github.com/CharJon/GeCO",
    license="MIT License",
    packages=find_packages(exclude=("tests", "docs", "data", "notebooks", "examples")),
    install_requires=["pyscipopt", "networkx", "numpy", "scipy", "pandas"],
)



from urllib.request import urlopen

FILES_BASE_URL = "http://people.brunel.ac.uk/~mastjjb/jeb/orlib/files/"


def read_number(line):
    if not line:
        return None
    return _str_to_number(line.strip().split(b" ")[0])


def read_numbers(line):
    return (_str_to_number(n) for n in line.strip().split(b" "))


def read_multiline_numbers(file, number_to_read):
    numbers = []
    while file:
        if len(numbers) == number_to_read:
            break
        elif len(numbers) > number_to_read:
            raise ValueError("Found more numbers than expected")
        else:
            line = file.readline()
            numbers_in_line = list(read_numbers(line))
            numbers += numbers_in_line
    return numbers


def _str_to_number(string):
    if b"." in string:
        return float(string)
    else:
        return int(string)


def zero_index(numbers):
    return map(lambda x: x - 1, numbers)


def orlib_load_instance(instance_name, reader, formulation):
    """
    Parameters
    ----------
    instance_name: str
        Name of instance file
    reader: function (file) -> params: tuple
        Takes a file-like object and returns the read parameters
    formulation: function (params: tuple) -> scip.model
        Takes a tuple of params and returns the generated model

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the loaded instance

    References
    ----------
    ..[1] http://people.brunel.ac.uk/~mastjjb/jeb/info.html
    """

    content_as_file = urlopen(FILES_BASE_URL + instance_name)
    params = reader(content_as_file)
    return formulation(*params)



import pytest

from geco.mips.max_cut.generic import *
from geco.mips.max_cut.tang import *


def test_tang():
    n, m = 100, 50
    model = tang_instance(n, m)
    assert model.getNVars() == m + n
    assert model.getNConss() == 2 * m
    assert model.getObjectiveSense() == "maximize"


def test_triangle():
    graph = nx.generators.cycle_graph(4)
    for _, _, data in graph.edges(data=True):
        data["weight"] = 1
    _, model = triangle(graph)
    n = len(graph.nodes)
    assert model.getNVars() == n * (n - 1) / 2
    assert (
        model.getNConss() == n * (n - 1) * (n - 2) / 3
    )  # 2 constraints for each triple of nodes
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 4


def test_naive_negative():
    graph = nx.generators.complete_graph(3)
    for _, _, data in graph.edges(data=True):
        data["weight"] = -1
    _, model = naive(graph)
    n, m = len(graph), len(graph.edges)
    assert model.getNVars() == n + m
    assert model.getNConss() == 4 * m
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 0


def test_naive_non_negative():
    graph = nx.generators.complete_graph(3)
    for _, _, data in graph.edges(data=True):
        data["weight"] = 1
    _, model = naive(graph)
    n, m = len(graph), len(graph.edges)
    assert model.getNVars() == n + m
    assert model.getNConss() == 2 * m
    model.hideOutput()
    model.optimize()
    assert model.getStatus() == "optimal"
    assert model.getObjVal() == 2


@pytest.mark.parametrize(
    "n,seed1,seed2",
    itertools.product([3, 10, 100], [0, 1, 1337, 53115], [0, 1, 1337, 53115]),
)
def test_seeding(n, seed1, seed2):
    graph = nx.generators.complete_graph(n)
    weights1 = tang_params(graph, seed=seed1)
    weights2 = tang_params(graph, seed=seed2)
    same_seeds_produce_same_params = seed1 == seed2 and weights1 == weights2
    different_seeds_produce_different_params = seed1 != seed2 and weights1 != weights2
    assert same_seeds_produce_same_params or different_seeds_produce_different_params



import networkx as nx
import pyscipopt as scip

from geco.mips.independent_set.generic import independent_set


def barabasi_albert_params(n, m, seed=0):
    """
    Generates a maximum independent set instance params of graphs described in [1].

    Parameters
    ----------
    n: int
        Number of nodes
    m: int
        Number of edges to attach from a new node to existing nodes
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    graph: nx.Graph
        A Barabasi-Albert Networkx graph

    References
    ----------
    .. [1] A. L. Barabási and R. Albert “Emergence of scaling in random networks”, Science 286, pp 509-512, 1999.
    """
    return nx.generators.barabasi_albert_graph(n, m, seed)


def barabasi_albert_instance(n, m, seed=0):
    """
    Generates a maximum independent set instance of graphs described in [1].

    Parameters
    ----------
    n: int
        Number of nodes
    m: int
        Number of edges to attach from a new node to existing nodes
    seed: integer, random_state, or None
        Indicator of random number generation state

    Returns
    -------
    model: scip.Model
        A pyscipopt model of the instance.

    References
    ----------
    .. [1] A. L. Barabási and R. Albert “Emergence of scaling in random networks”, Science 286, pp 509-512, 1999.
    """
    return independent_set(
        barabasi_albert_params(n, m, seed), name="Barabasi-Albert Independent Set"
    )



import itertools
import math

import pyscipopt as scip
from geco.mips.scheduling.generic import generate_params
from networkx.utils import py_random_state
from geco.mips.scheduling.generic import *


@py_random_state(-1)
def hooker_params(number_of_facilities, number_of_tasks, seed=0):
    """Generates late tasks mip instance described in section 4 in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
    processing_times: dict[int,int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[int,int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int, float]
        deadline (time step) to finish a job

    References
    ----------
    .. [1] Hooker, John. (2005). Planning and Scheduling to Minimize
     Tardiness. 314-327. 10.1007/11564751_25.
    """
    return generate_params(number_of_facilities, number_of_tasks, seed)[:-1]


@py_random_state(-1)
def hooker_instance(number_of_facilities, number_of_tasks, time_steps, seed=0):
    """Generates late tasks mip instance described in section 4 in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    time_steps:
        the number of time steps starting from 0 (corresponds to "N" in the paper)

    Returns
    -------
        model: SCIP model of the late tasks instance

    References
    ----------
    .. [1] Hooker, John. (2005). Planning and Scheduling to Minimize
     Tardiness. 314-327. 10.1007/11564751_25.
    """
    return late_tasks_formulation(
        number_of_facilities,
        number_of_tasks,
        time_steps,
        *hooker_params(number_of_facilities, number_of_tasks, seed),
        name="Hooker Scheduling Instance",
    )


def generate_hookers_instances():
    number_of_tasks = [10 + 2 * i for i in range(7)]
    time_steps = [10, 100]
    seeds = range(10)
    for n, t, seed in itertools.product(number_of_tasks, time_steps, seeds):
        params = 3, n, t, seed
        yield params, hooker_instance(*params)


def _common_hooker_params(number_of_facilities, number_of_tasks, seed):
    capacities = [10] * number_of_facilities
    resource_requirements = {}
    for i in range(number_of_tasks):
        cur_res_requirement = seed.randrange(1, 10)
        for j in range(number_of_facilities):
            resource_requirements[i, j] = cur_res_requirement
    return capacities, resource_requirements


@py_random_state(-1)
def c_instance_params(seed=0):
    for m, n in itertools.product(range(2, 4 + 1), range(10, 38 + 1, 2)):
        yield c_params_generator(m, n, seed)


@py_random_state(-1)
def c_params_generator(number_of_facilities, number_of_tasks, seed=0):
    """
    Generate instance parameters for the c problem set mentioned in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int, int]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility

    References
    ----------
    ..[1] http://public.tepper.cmu.edu/jnh/instances.htm
    """
    capacities, resource_requirements = _common_hooker_params(
        number_of_facilities, number_of_tasks, seed
    )

    release_dates = [0] * number_of_tasks
    due_dates = [
        _due_date_helper(1 / 3, number_of_facilities, number_of_tasks)
    ] * number_of_tasks

    processing_times = {}
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_times[j, i] = seed.randrange(i + 1, 10 * (i + 1))

    processing_costs = {}
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_costs[j, i] = seed.randrange(
                2 * (number_of_facilities - i), 20 * (number_of_facilities - i)
            )

    return (
        number_of_facilities,
        number_of_tasks,
        processing_times,
        capacities,
        processing_costs,
        release_dates,
        due_dates,
        resource_requirements,
    )


@py_random_state(-1)
def e_instance_params(seed=0):
    for m in range(2, 10 + 1):
        yield e_params_generator(m, 5 * m, seed)


@py_random_state(-1)
def e_params_generator(number_of_facilities, number_of_tasks, seed=0):
    """
    Generate instance parameters for the e problem set mentioned in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int, int]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility

    References
    ----------
    ..[1] http://public.tepper.cmu.edu/jnh/instances.htm
    """
    capacities, resource_requirements = _common_hooker_params(
        number_of_facilities, number_of_tasks, seed
    )

    release_dates = [0] * number_of_tasks
    due_dates = [33] * number_of_tasks

    processing_times = {}
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_times[j, i] = seed.randrange(
                2, int(25 - i * (10 / (number_of_facilities - 1)))
            )

    processing_costs = {}
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_costs[j, i] = seed.randrange(
                math.floor(400 / (25 - i * (10 / (number_of_facilities - 1)))),
                math.ceil(800 / (25 - i * (10 / (number_of_facilities - 1)))),
            )

    return (
        number_of_facilities,
        number_of_tasks,
        processing_times,
        capacities,
        processing_costs,
        release_dates,
        due_dates,
        resource_requirements,
    )


@py_random_state(-1)
def de_instance_params(seed=0):
    for n in range(14, 28 + 1, 2):
        yield de_params_generator(3, n, seed)


@py_random_state(-1)
def de_params_generator(number_of_facilities, number_of_tasks, seed=0):
    """
    Generate instance parameters for the de problem set mentioned in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int, int]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility

    References
    ----------
    ..[1] http://public.tepper.cmu.edu/jnh/instances.htm
    """
    capacities, resource_requirements = _common_hooker_params(
        number_of_facilities, number_of_tasks, seed
    )

    release_dates = [0] * number_of_tasks
    due_dates = [
        seed.randrange(
            _due_date_helper((1 / 4) * (1 / 3), number_of_facilities, number_of_tasks),
            _due_date_helper(1 / 3, number_of_facilities, number_of_tasks),
        )
        for _ in range(number_of_tasks)
    ]

    processing_times = {}
    range_start = 2 if number_of_facilities <= 20 else 5  # P1 in the reference website
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_times[j, i] = seed.randrange(range_start, 30 - i * 5)

    processing_costs = {}
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_costs[j, i] = seed.randrange(10 + 10 * i, 40 + 10 * i)

    return (
        number_of_facilities,
        number_of_tasks,
        processing_times,
        capacities,
        processing_costs,
        release_dates,
        due_dates,
        resource_requirements,
    )


@py_random_state(-1)
def df_instance_params(seed=0):
    for n in range(14, 28 + 1, 2):
        yield df_params_generator(3, n, seed)


@py_random_state(-1)
def df_params_generator(number_of_facilities, number_of_tasks, seed=0):
    """
    Generate instance parameters for the df problem set mentioned in [1].

    Parameters
    ----------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    seed: int, random object or None
        for randomization

    Returns
    -------
    number_of_facilities: int
        the number of facilities to schedule on
    number_of_tasks: int
        the number of tasks to assign to facilities
    processing_times: dict[(int,int),int]
        time steps to process each task
    capacities: list[int]
        capacity of each facility
    assignment_costs: dict[(int,int),int]
        cost of assigning a task to a facility
    release_times: list[int]
        time step at which a job is released
    deadlines: dict[int, int]
        deadline (time step) to finish a job
    resource_requirements: dict[(int,int),int]
        resources required for each task assigned to a facility

    References
    ----------
    ..[1] http://public.tepper.cmu.edu/jnh/instances.htm
    """
    capacities, resource_requirements = _common_hooker_params(
        number_of_facilities, number_of_tasks, seed
    )

    release_dates = [0] * number_of_tasks

    random_release_time = seed.choice(release_dates)
    due_dates = [
        seed.randrange(
            random_release_time
            + _due_date_helper(1 / 4 * 1 / 2, number_of_facilities, number_of_tasks),
            random_release_time
            + _due_date_helper(1 / 2, number_of_facilities, number_of_tasks),
        )
        for _ in range(number_of_tasks)
    ]

    processing_times = {}
    range_start = 2 if number_of_facilities <= 20 else 5  # P1 in the reference website
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_times[j, i] = seed.randrange(range_start, 30 - i * 5)

    processing_costs = {}
    for i in range(number_of_facilities):
        for j in range(number_of_tasks):
            processing_costs[j, i] = seed.randrange(10 + 10 * i, 40 + 10 * i)

    return (
        number_of_facilities,
        number_of_tasks,
        processing_times,
        capacities,
        processing_costs,
        release_dates,
        due_dates,
        resource_requirements,
    )


def _due_date_helper(a, number_of_facilities, number_of_tasks):
    return math.ceil(
        5 * a * number_of_tasks * (number_of_facilities + 1) / number_of_facilities
    )
